from collections.abc import Callable
import functools
import logging
import os
import re
import warnings

import numpy as np
import xarray as xr
from pyaerocom import const
from pyaerocom.exceptions import VarNotAvailableError
from pyaerocom.griddeddata import GriddedData
from pyaerocom.io.gridded_reader import GriddedReader
from pyaerocom.projection_information import ProjectionInformation
from pyaerocom.units.helpers import get_standard_unit
from pyaerocom.units.units import Unit

from .additional_variables import (
    add_dataarrays,
    calc_concNhno3,
    calc_concNnh3,
    calc_concNnh4,
    calc_concNno,
    calc_concNno2,
    calc_concNno3pm10,
    calc_concNno3pm25,
    calc_concno3pm10,
    calc_concno3pm25,
    calc_concso4t,
    calc_concSso2,
    calc_concsspm25,
    calc_conNtnh_emep,
    calc_conNtno3_emep,
    calc_vmrno2,
    calc_vmro3,
    calc_vmrox_from_conc,
    identity,
    subtract_dataarrays,
    update_EC_units,
    calc_ratpm10pm25,
    calc_ratpm25pm10,
)
from .model_variables import emep_variables
import pathlib

logger = logging.getLogger(__name__)


class ReadMscwCtm(GriddedReader):
    """
    Class for reading model output from the EMEP MSC-W chemical transport model.

    Parameters
    ----------
    data_id : str
        string ID of model (e.g. "AATSR_SU_v4.3","CAM5.3-Oslo_CTRL2016")
    data_dir : str, optional
        Base directory of EMEP data, containing one or more netcdf files.
        data_dir must contain at least one of Base_(hour|day|month|fullrun).nc
        For multi-year analysis/trends, datadir may contain subdirectories named by trend-year,
        e.g. 2010 or trend2010.
    file_pattern : str, optional
        Optional regular expression against which the base name of files will be matched.
        This can be used to override the default `Base_{freq}.nc` file matching.

        Note that for convenience the string literal '{freq}' can be included as part of the
        pattern and will be expanded to (hour|day|month|fullrun). This is recommended, as
        the presence of these strings are used to derive ts_type, which is currently necessary
        for reading.

    Attributes
    ----------
    data_id : str
        ID of model
    """

    # dictionary containing information about additionally required variables
    # for each auxiliary variable (i.e. each variable that is not provided
    # by the original data but computed on import)
    AUX_REQUIRES = {
        "depso4": ["dryso4", "wetso4"],
        "depoxs": ["dryoxs", "wetoxs"],
        "depoxn": ["dryoxn", "wetoxn"],
        "deprdn": ["dryrdn", "wetrdn"],
        "concbc": ["concbcf", "concbcc"],
        "concno3": ["concno3c", "concno3f"],
        "concoa": ["concoac", "concoaf"],
        "concpmgt25": ["concpm10", "concpm25"],
        "concNhno3": ["conchno3"],
        "concNnh3": ["concnh3"],
        "concNnh4": ["concnh4"],
        "concNno3pm10": ["concno3f", "concno3c"],
        "concNno3pm25": ["concno3f", "concno3c"],
        "concno3pm10": ["concno3f", "concno3c"],
        "concno3pm25": ["concno3f", "concno3c"],
        "concsspm25": ["concssf", "concssc"],
        "concsspm10": ["concssf", "concssc"],
        # "vmrox": ["concno2", "vmro3"],
        "vmrox": ["concno2", "conco3"],
        "vmrno2": ["concno2"],
        "concNtno3": ["concoxn"],
        "concNtnh": ["concrdn"],
        # "concNtno3": ["conchno3", "concno3f", "concno3c"],
        # "concNtnh": ["concnh3", "concnh4"],
        "concecpm25": ["concecFine"],
        "concecpm10": ["concecFine", "concecCoarse"],
        "concCecpm10": ["concecpm10"],
        "concCecpm25": ["concecpm25"],
        "concCocpm25": ["concCocFine"],
        "concCocpm10": ["concCocFine", "concCocCoarse"],
        "concso4t": ["concso4", "concss"],
        "concNno": ["concno"],
        "concNno2": ["concno2"],
        "concSso2": ["concso2"],
        "vmro3": ["conco3"],
        "ratpm10pm25": ["concpm10", "concpm25"],
        "ratpm25pm10": ["concpm25", "concpm10"],
        # For Pollen
        # "concpolyol": ["concspores"],
        # For EC
        "concecFineRes": ["concecFineResNew", "concecFineResAge"],
        "concecFineNonRes": ["concecFineNonResNew", "concecFineNonResAge"],
        "concecTotalRes": ["concecFineRes", "concecCoarseRes"],
        "concecTotalNonRes": ["concecFineNonRes", "concecCoarseNonRes"],
        "concebc": ["concecFine", "concecCoarse"],
        # For EC from emission
        "concecTotalResEM": ["concecFineResNewEM", "concecFineResAgeEM"],
        "concecTotalNonResEM": ["concecFineNonResNewEM", "concecFineNonResAgeEM"],
        "concebcem": ["concecFineEM", "concecCoarseEM"],
        "concCecpm25EM": ["concecFineEM"],
        "concCecpm10EM": ["concecFineEM"],
    }

    # Functions that are used to compute additional variables (i.e. one
    # for each variable defined in AUX_REQUIRES)
    # NOTE: these methods are supposed to work for xarray.DataArray instances
    # not iris.cube.Cube instance
    AUX_FUNS = {
        "depso4": add_dataarrays,
        "depoxs": add_dataarrays,
        "depoxn": add_dataarrays,
        "deprdn": add_dataarrays,
        "concbc": add_dataarrays,
        "concno3": add_dataarrays,
        "concoa": add_dataarrays,
        "concpmgt25": subtract_dataarrays,
        "concNhno3": calc_concNhno3,
        "concNnh3": calc_concNnh3,
        "concNnh4": calc_concNnh4,
        "concNno3pm10": calc_concNno3pm10,
        "concNno3pm25": calc_concNno3pm25,
        "concno3pm10": calc_concno3pm10,
        "concno3pm25": calc_concno3pm25,
        "concsspm25": calc_concsspm25,
        "concsspm10": add_dataarrays,
        # "vmrox": calc_vmrox,
        "vmrox": calc_vmrox_from_conc,
        "vmrno2": calc_vmrno2,
        "concNtno3": calc_conNtno3_emep,
        "concNtnh": calc_conNtnh_emep,
        # "concNtno3": calc_conNtno3,
        # "concNtnh": calc_concNtnh,
        "concecpm25": identity,
        "concecpm10": add_dataarrays,
        "concCecpm25": update_EC_units,
        "concCecpm10": update_EC_units,
        "concCocpm25": identity,
        "concCocpm10": add_dataarrays,
        "concso4t": calc_concso4t,
        "concNno": calc_concNno,
        "concNno2": calc_concNno2,
        "concSso2": calc_concSso2,
        "vmro3": calc_vmro3,
        "ratpm10pm25": calc_ratpm10pm25,
        "ratpm25pm10": calc_ratpm25pm10,
        # "concpolyol": calc_concpolyol,
        # For EC
        "concecFineRes": add_dataarrays,
        "concecFineNonRes": add_dataarrays,
        "concecTotalRes": add_dataarrays,
        "concecTotalNonRes": add_dataarrays,
        "concebc": add_dataarrays,
        # For EC from emission
        "concecTotalResEM": add_dataarrays,
        "concecTotalNonResEM": add_dataarrays,
        "concebcem": add_dataarrays,
        "concCecpm25EM": update_EC_units,
        "concCecpm10EM": update_EC_units,
    }

    #: supported filename template, freq-placeholder is for frequencies
    FILE_FREQ_TEMPLATE = "Base_{freq}.nc"

    #: frequencies encoded in filenames
    FREQ_CODES = {
        "hour": "hourly",
        "day": "daily",
        "month": "monthly",
        "fullrun": "yearly",
    }

    REVERSE_FREQ_CODES = {
        "hourly": "hour",
        "daily": "day",
        "monthly": "month",
        "yearly": "fullrun",
    }

    DEFAULT_FILE_NAME = "Base_day.nc"

    #: pattern for 4-digit years for 19XX and 20XX used for trend subdirectories
    YEAR_PATTERN = r".*((?:19|20)\d\d).*"

    class _PrivateFields:
        filename: str | None = None
        filedata: xr.Dataset | None = None
        filepaths: list[str] | None = None
        files: list[str] | None = None
        data_dir: str | None = None
        file_pattern: re.Pattern
        ts_type: str | None = None

    def __init__(
        self,
        data_id: str | None = None,
        data_dir: str | None = None,
        *,
        file_pattern: str | None = None,
        **kwargs,
    ):
        # opened dataset (for performance boost), will be reset if data_dir is
        # changed
        self._private = self._PrivateFields()

        self.var_map = emep_variables()
        if "emep_vars" in kwargs:
            new_map = kwargs["emep_vars"]
            if isinstance(new_map, dict):
                self.var_map.update(new_map)
            else:
                logger.warning(f"New map {new_map} is not a dict. Skipping")

        if file_pattern is None:
            # Pattern for the 'Base_{freq}.nc' default strategy.
            file_pattern = rf"^Base_({'|'.join(self.FREQ_CODES.keys())}).nc$"
        elif isinstance(file_pattern, str):
            file_pattern = file_pattern.format(freq=f"({'|'.join(self.FREQ_CODES.keys())})")
        else:
            raise TypeError(
                f"file_pattern should be of type str or None. Got {type(file_pattern)}"
            )

        try:
            file_pattern = re.compile(file_pattern)
        except re.error as e:
            raise ValueError(
                f"Provided file_pattern '{file_pattern}' can't be compiled to re.Pattern."
            ) from e

        self._private.file_pattern = file_pattern
        logger.info(
            f"Matching valid EMEP files based on the following regular expression: '{file_pattern.pattern}'"
        )

        if data_dir is not None:
            if not isinstance(data_dir, str) or not os.path.exists(data_dir):
                raise FileNotFoundError(f"{data_dir}")

            self._data_dir = data_dir

        self._data_id = data_id
        self._private.filename = self.DEFAULT_FILE_NAME

    def _search_all_files(self):
        folders = self._get_trend_folders_from_folder()
        self._filepaths = self._get_files_from_folders(folders)

    def _get_files_from_folders(self, folders: list[str]):
        files: list[str] = []
        for d in folders:
            files += self._check_files_in_data_dir(d)
        return files

    def _get_trend_folders_from_folder(self):
        """
        Finds all the subfolders where a emep file for one year might be.

        Note
        -------
        Checks only current level for folders. Should be able
        to search deeper.

        The only qualification of being a valid subfolder is whether or not
        the subfolder has contains a number >= 1900 < 2100. There are checks if there
        are emep files in the folder

        Raises
        ------
        FileNotFoundError if no mscw-files can be found in data-dir

        Returns
        -------
        List
         List of the names of the subfolder

        """
        dd = self._data_dir

        folders: list[str] = []
        yrs: list[int] = []
        for d in os.listdir(dd):
            if not os.path.isdir(os.path.join(dd, d)):
                continue
            m = re.match(self.YEAR_PATTERN, d)
            if m is not None:
                has_mscwfiles = False
                for f in os.listdir(os.path.join(dd, d)):
                    if (
                        os.path.isfile(os.path.join(dd, d, f))
                        and self._private.file_pattern.match(f) is not None
                    ):
                        has_mscwfiles = True

                if has_mscwfiles:
                    yrs.append(int(m.group(1)))
                    folders.append(os.path.join(dd, d))

        if len(folders) == 0:  # no trends, use folder
            for f in os.listdir(dd):
                if os.path.isfile(os.path.join(dd, f)) and self._private.file_pattern.match(f):
                    folders = [dd]
                    break

            if len(folders) == 0:
                raise FileNotFoundError(
                    f"no files matching {self._private.file_pattern} found in {dd}"
                )
        else:
            folders = [d for _, d in sorted(zip(yrs, folders))]
        return list(set(folders))

    @staticmethod
    @functools.cache
    def _get_year_from_nc(filename: str) -> int:
        with xr.open_dataset(filename, decode_timedelta=True) as nc:
            return np.mean(nc["time"][:]).data.astype("datetime64[Y]").astype(int) + 1970

    def _get_yrs_from_filepaths(self) -> list[str]:
        """Get available years of data from the filepaths. The year of the first
        Base_*.nc dataset in the filepath is read from the time-variable of the nc-file.

        :return: list of years as str
        """
        fps = self._filepaths
        yrs = []
        for fp in fps:
            try:
                yr = ReadMscwCtm._get_year_from_nc(fp)
            except Exception as ex:
                raise ValueError(f"Could not find any year in {fp}: {ex}")
            yrs.append(str(yr))

        return sorted(list(set(yrs)))

    def _get_tst_from_file(self, file: str):
        _, fname = os.path.split(file)

        # Note: This is to maintain previous functionality which would raise error if file did not match
        # Base_{freq} template. I am not sure if this should be the responsibility of this function, and
        # alternatively this can be removed (including the test).
        if self._private.file_pattern.match(file) is None:
            raise ValueError(
                f"The file '{file}' does not match file_pattern '{self._private.file_pattern}'"
            )

        for freq, tst in self.FREQ_CODES.items():
            if freq in fname:
                return tst

    def _clean_filepaths(self, filepaths: list[str], yrs: list[str], ts_type: str):
        clean_paths: set[str] = set()
        found_yrs: set[str] = set()

        yrs = [int(yr) for yr in yrs]
        for path in filepaths:
            file = os.path.split(path)[1]

            if self._get_tst_from_file(file) != ts_type:
                logger.debug(f"ignoring file {path}: not of type {ts_type}")
                continue

            try:
                yr = ReadMscwCtm._get_year_from_nc(path)
            except Exception as ex:
                raise ValueError(f"Could not find any year in {path}: {ex}")

            clean_paths.add(path)
            if yr not in yrs:
                raise ValueError(f"The year {yr} of {path} is not in {yrs}")

            if yr in found_yrs:
                continue

            found_yrs.add(yr)

        if len(found_yrs) != len(yrs):
            raise ValueError(
                f"A different amount of years {found_yrs} were found compared to {yrs} in {filepaths}"
            )

        return list(clean_paths)

    @property
    def data_id(self) -> str | None:
        return self._data_id

    @property
    def _data_dir(self) -> str:
        """
        Directory containing netcdf files
        """
        if self._private.data_dir is None:
            raise AttributeError("data_dir needs to be set before accessing")
        return self._private.data_dir

    @_data_dir.setter
    def _data_dir(self, val: str):
        if val is None:
            raise ValueError(f"Data dir {val} needs to be a dictionary or a file")
        if not os.path.isdir(val):
            raise FileNotFoundError(val)
        self._private.data_dir = val
        self._private.filedata = None
        self._search_all_files()
        self._private.files = self._filepaths

    @property
    def _filename(self) -> str | None:
        """
        Name of latest netcdf file read
        """
        return self._private.filename

    @_filename.setter
    def _filename(self, val: str):
        """
        Name of netcdf file
        """
        if not isinstance(val, str):  # pragma: no cover
            raise ValueError("needs str")
        elif val == self._private.filename:
            return
        self._private.filename = val
        self._private.filedata = None

    @property
    def _filepaths(self) -> list[str]:
        """
        Paths to data file
        """
        if self._data_dir is None and self._filepaths is None:  # pragma: no cover
            raise AttributeError("data_dir or filepaths needs to be set before accessing")
        return self._private.filepaths

    @_filepaths.setter
    def _filepaths(self, value: list[str]):
        if not isinstance(value, list):  # pragma: no cover
            raise ValueError("needs to be list of strings")
        self._private.filepaths = value

    @property
    def _filedata(self) -> xr.Dataset:
        """
        Loaded netcdf file (:class:`xarray.Dataset`)
        """
        if self._private.filedata is None:
            self._open_file()
        return self._private.filedata

    @functools.cache
    def _check_files_in_data_dir(self, data_dir: str):
        """
        Check for data files in input data directory

        Parameters
        ----------
        data_dir : str
            directory to be searched.

        Raises
        ------
        FileNotFoundError
            if no EMEP files can be identified

        Returns
        -------
        list
            list of file matches

        """
        files: list[str] = [str(p) for p in pathlib.Path(data_dir).iterdir() if p.is_file()]

        matches = [
            str(p)
            for p in pathlib.Path(data_dir).iterdir()
            if self._private.file_pattern.match(p.name) is not None
        ]
        if len(matches) == 0:
            raise FileNotFoundError(
                f"No valid model files could be found in {data_dir} for any of the "
                f"supported files: {files}"
            )
        return matches

    @property
    def _ts_type(self) -> str:
        """
        Frequency of time dimension of current data file

        Raises
        ------
        AttributeError
            if :attr:`filename` is not set.

        Returns
        -------
        str
            current ts_type.

        """
        if self._private.ts_type is None:
            return self._ts_type_from_filename(self._filename)
        return self._private.ts_type

    @property
    def ts_types(self) -> list[str]:
        """
        List of available frequencies

        Raises
        ------
        AttributeError
            if :attr:`data_dir` is not set.

        Returns
        -------
        list
            list of available frequencies

        """
        if not isinstance(self._private.files, list):
            raise AttributeError("please set data_dir first")
        tsts = []
        for file in self._private.files:
            tsts.append(self._ts_type_from_filename(file))
        return list(set(tsts))

    @property
    def years_avail(self) -> list[str]:
        """
        Years available in loaded dataset
        """
        years = self._get_yrs_from_filepaths()

        years = list(np.unique(years))
        return sorted(years)

    @property
    def vars_provided(self) -> list[str]:
        """Variables provided by this dataset"""
        return list(self.var_map) + list(self.AUX_REQUIRES)

    def _open_file(self):
        """
        Open current netcdf file

        Returns
        -------
        dict(xarray.Dataset)
            Dict with years as keys and Datasets as items

        """
        fps = self._filepaths
        ds = {}
        yrs = self._get_yrs_from_filepaths()

        ts_type = self._ts_type
        fps = self._clean_filepaths(fps, yrs, ts_type)

        if ts_type == "hourly" and len(fps) > 1:
            start_date = None
            end_date = None
            for fp in fps:
                with xr.open_dataset(fp, decode_timedelta=True) as nc:
                    file_start_date = nc["time"][:].data.min()
                    file_end_date = nc["time"][:].data.max()

                start_date = min([x for x in [start_date, file_start_date] if x is not None])
                end_date = max([x for x in [end_date, file_end_date] if x is not None])

            if (end_date - start_date) / np.timedelta64(1, "h") > (366 * 24):
                raise ValueError(
                    f"ts_type {ts_type} can not be hourly when using multiple years ({start_date} - {end_date})"
                )

        logger.info(f"Opening {fps}")
        ds = xr.open_mfdataset(fps, chunks={"time": 24}, decode_timedelta=True)

        self._private.filedata = ds

        return ds

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "ReadMscwCtm"

    def has_var(self, var_name: str) -> bool:
        """Check if variable is supported

        Parameters
        ----------
        var_name : str
            variable to be checked

        Returns
        -------
        bool
        """
        avail = self.vars_provided
        if var_name in avail or const.VARS[var_name].var_name_aerocom in avail:
            return True
        return False

    def _ts_type_from_filename(self, filename: str) -> str:
        """
        Get ts_type from filename

        Parameters
        ----------
        filename : str

        Raises
        ------
        ValueError
            if ts_type cannot be inferred from filename.

        Returns
        -------
        tstype : str
        """
        filename = os.path.basename(filename)
        for substr, tstype in self.FREQ_CODES.items():
            if substr in filename:
                return tstype
        raise ValueError(f"Failed to retrieve ts_type from filename {filename}")

    def _filename_from_ts_type(self, ts_type: str):
        """
        Infer file name of data based on input ts_type

        Parameters
        ----------
        ts_type : str
            desired time freq of data

        Raises
        ------
        ValueError
            On wrong ts_type.

        Returns
        -------
        fname : str
            Name of data file based on ts_type
            freq.

        """
        if ts_type not in self.REVERSE_FREQ_CODES:
            raise ValueError(f"unknown ts_type={ts_type}")
        freq = self.REVERSE_FREQ_CODES[ts_type]
        return self.FILE_FREQ_TEMPLATE.format(freq=freq)

    def _compute_var(self, var_name_aerocom: str, ts_type: str):
        """Compute auxiliary variable

        Like :func:`read_var` but for auxiliary variables
        (cf. AUX_REQUIRES)

        Parameters
        ----------
        var_name : str
            variable that are supposed to be read
        ts_type : str
            string specifying temporal resolution.

        Returns
        -------
        xarray.DataArray
            loaded data object

        """

        temp_arrs = []
        req = self.AUX_REQUIRES[var_name_aerocom]
        aux_func = self.AUX_FUNS[var_name_aerocom]
        logger.info(f"computing {var_name_aerocom} from {req} using {aux_func}")
        proj_info = None
        for aux_var in self.AUX_REQUIRES[var_name_aerocom]:
            arr, proj_info = self._load_var(aux_var, ts_type)
            temp_arrs.append(arr)

        return aux_func(*temp_arrs), proj_info

    def _load_var(self, var_name_aerocom: str, ts_type: str):
        """
        Load variable data as :class:`xarray.DataArray`.

        This combines both, variables that can be read directly and auxiliary
        variables that are computed.

        Parameters
        ----------
        var_name_aerocom : str
            variable name
        ts_type : str
            desired frequency

        Raises
        ------
        VarNotAvailableError
            if input variable is not available

        Returns
        -------
        xarray.DataArray
            loaded data
        ProjectionInformation
            projection of variable

        """
        if var_name_aerocom in self.var_map:  # can be read
            return self._read_var_from_file(var_name_aerocom, ts_type)
        elif var_name_aerocom in self.AUX_REQUIRES:
            return self._compute_var(var_name_aerocom, ts_type)
        raise VarNotAvailableError(
            f"Variable {var_name_aerocom} is not supported"
        )  # pragma: no cover

    def read_var(self, var_name: str, ts_type: str | None = None, **kwargs):
        """Load data for given variable.

        Parameters
        ----------
        var_name : str
            Variable to be read
        ts_type : str
            Temporal resolution of data to read. Supported are
            "hourly", "daily", "monthly" , "yearly".

        Returns
        -------
        GriddedData
        """
        if not self.has_var(var_name):
            raise VarNotAvailableError(var_name)
        var = const.VARS[var_name]
        var_name_aerocom = var.var_name_aerocom

        if self._data_dir is None:  # pragma: no cover
            raise ValueError("data_dir must be set before reading.")
        elif self._filename is None and ts_type is None:  # pragma: no cover
            raise ValueError("please specify ts_type")
        elif ts_type is not None:
            # filename and ts_type are set. update filename if ts_type suggests
            # that current file has different resolution
            self._filename = self._filename_from_ts_type(ts_type)
        self._private.ts_type = ts_type

        ts_type = self._ts_type

        arr, proj_info = self._load_var(var_name_aerocom, ts_type)
        if arr.units in Unit._UALIASES:
            arr.attrs["units"] = Unit._UALIASES[arr.units]
        try:
            cube = arr.to_iris()
        except MemoryError as e:  # pragma: no cover
            raise NotImplementedError from e

        if ts_type == "hourly":
            cube.coord("time").convert_units("hours since 1900-01-01")
        gridded = GriddedData(
            cube,
            var_name=var_name_aerocom,
            ts_type=ts_type,
            check_unit=True,
            convert_unit_on_init=True,
            proj_info=proj_info,
        )

        # At this point a GriddedData object with name gridded should exist

        gridded.metadata["data_id"] = self._data_id
        gridded.metadata["from_files"] = self._filepaths

        gridded.convert_unit(get_standard_unit(var_name))
        # Remove unnecessary metadata. Better way to do this?
        for metadata in ["current_date_first", "current_date_last"]:
            if metadata in gridded.metadata.keys():
                del gridded.metadata[metadata]
        return gridded

    def _read_var_from_file(self, var_name_aerocom: str, ts_type: str):
        """
        Read variable data from file as :class:`xarray.DataArray`.

        See also :func:`_load_var`

        Parameters
        ----------
        var_name_aerocom : str
            variable name
        ts_type : str
            desired frequency

        Raises
        ------
        VarNotAvailableError
            if input variable is not available

        Returns
        -------
        xarray.DataArray
            loaded data
        ProjectionInformation
            projection of variable

        """
        emep_var = self.var_map[var_name_aerocom]

        try:
            filedata = self._filedata
            data = filedata[emep_var]
            proj_info = ProjectionInformation.from_xarray(filedata, emep_var)
        except KeyError:
            raise VarNotAvailableError(
                f"{var_name_aerocom} ({emep_var}) not available in {self._filename}"
            )
        data.attrs["long_name"] = var_name_aerocom
        data.time.attrs["long_name"] = "time"
        data.time.attrs["standard_name"] = "time"
        prefix = emep_var.split("_")[0]
        data.attrs["units"] = self._preprocess_units(data.units, prefix)
        return data, proj_info

    @staticmethod
    def _preprocess_units(units: str, prefix: str | None = None):
        """
        Update units for certain variables

        Parameters
        ----------
        units : str
            Current unit of data
        prefix : str, optional
            Variable prefix (e.g. AOD, AbsCoeff).

        Returns
        -------
        str
            updated unit (where applicable)

        """
        if units == "" and prefix == "AOD":  #
            return "1"
        elif units == "" and prefix == "AbsCoef":
            return "m-1"
        return units

    def add_aux_compute(self, var_name: str, vars_required: list[str] | str, fun: Callable):
        """Register new variable to be computed

        Parameters
        ----------
        var_name : str
            variable name to be computed
        vars_required : list
            list of variables to read, that are required to compute `var_name`
        fun : callable
            function that takes a list of `GriddedData` objects as input and
            that are read using variable names specified by `vars_required`.
        """
        if isinstance(vars_required, str):
            vars_required = [vars_required]
        if not isinstance(vars_required, list):
            raise ValueError(
                f"Invalid input for vars_required. Need str or list. Got: {vars_required}"
            )
        elif not callable(fun):
            raise ValueError("Invalid input for fun. Input is not a callable object")
        self.AUX_REQUIRES[var_name] = vars_required
        self.AUX_FUNS[var_name] = fun


class ReadEMEP(ReadMscwCtm):
    """Old name of :class:`ReadMscwCtm`."""

    def __init__(self, *args, **kwargs):
        warnings.warn(
            "You are using a deprecated name ReadEMEP for class ReadMscwCtm, "
            "please use ReadMscwCtm instead",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(*args, **kwargs)



import numpy as np
import numpy.typing as npt

from pyaerocom._lowlevel_helpers import BrowseDict


class VerticalProfile:
    """Object representing single variable profile data"""

    def __init__(
        self,
        data: npt.ArrayLike,
        altitude: npt.ArrayLike,
        dtime,
        var_name: str,
        data_err: npt.ArrayLike | None,
        var_unit: str,
        altitude_unit: str,
    ):
        self.var_name = var_name
        self.dtime = dtime
        self.data = data
        self.data_err = data_err
        self.altitude = altitude

        self.var_info = BrowseDict()
        self.var_info["altitude"] = dict(units=altitude_unit)
        self.var_info[self.var_name] = dict(units=var_unit)

        # Guard against having data (and data errors) with missing associated altitude info
        if hasattr(self.data_err, "__len__"):
            if not self.data.shape[-1] == self.altitude.shape[-1] == self.data_err.shape[-1]:
                raise ValueError(
                    "Data, data errors, and altitude arrays must have the same length."
                )
        else:
            if not self.data.shape[-1] == self.altitude.shape[-1]:
                raise ValueError(
                    "Data and altitude arrays must have the same length along the last axis."
                )

    @property
    def data(self):
        """Array containing data values corresponding to data"""
        return self._data

    @data.setter
    def data(self, val):
        if isinstance(val, list):
            val = np.asarray(val)
        self._data = val

    @property
    def data_err(self):
        """Array containing data values corresponding to data"""
        return self._data_err

    @data_err.setter
    def data_err(self, val):
        if isinstance(val, list):
            val = np.asarray(val)
        self._data_err = val

    @property
    def altitude(self):
        """Array containing altitude values corresponding to data"""
        return self._altitude

    @altitude.setter
    def altitude(self, val):
        if isinstance(val, list):
            val = np.asarray(val)
        self._altitude = val



from __future__ import annotations

from datetime import date
from pathlib import Path

import pytest
from typer.testing import CliRunner

from pyaerocom.io.cams2_83.models import ModelName, RunType
from pyaerocom.scripts.cams2_83.cli import app, make_config
from pyaerocom.scripts.cams2_83.evaluation import EvalType

runner = CliRunner()


@pytest.mark.usefixtures("fake_ExperimentProcessor", "reset_cachedir")
def test_clearcache(
    monkeypatch,
    fake_cache_path: Path,
    tmp_path: Path,
    caplog,
):
    assert list(fake_cache_path.glob("*.pkl"))

    options = f"forecast week 2024-03-16 2024-03-23 --model-path {tmp_path} --obs-path {tmp_path} --data-path {tmp_path} --coldata-path {tmp_path} --cache {fake_cache_path} --name 'Test'"
    result = runner.invoke(app, options.split())
    assert "Running Statistics" in caplog.text
    assert result.exit_code == 0
    # Check that the cache is cleared
    assert not list(fake_cache_path.glob("*.pkl"))


@pytest.mark.usefixtures("fake_CAMS2_83_Processer", "reset_cachedir")
def test_not_cleared_cache(
    monkeypatch,
    fake_cache_path: Path,
    tmp_path: Path,
    caplog,
):
    assert list(fake_cache_path.glob("*.pkl"))

    options = f"forecast long 2024-03-16 2024-03-23 --model-path {tmp_path} --obs-path {tmp_path} --data-path {tmp_path} --coldata-path {tmp_path} --name 'Test' --medianscores"
    result = runner.invoke(app, options.split())
    assert "Running CAMS2_83 Specific Statistics, cache is not cleared" in caplog.text
    assert result.exit_code == 0
    # Check that the cache is not cleared
    assert list(fake_cache_path.glob("*.pkl"))


def test_eval_dummy(
    tmp_path: Path,
    caplog,
):
    options = f"forecast day 2024-03-16 2024-03-16 --model-path {tmp_path} --obs-path {tmp_path} --data-path {tmp_path} --coldata-path {tmp_path} --name 'Test'"
    result = runner.invoke(app, options.split())
    assert result.exit_code == 0
    assert "Failed to read model variable" in caplog.text


def test_eval_medianscores_dummy(
    tmp_path: Path,
    caplog,
):
    options = f"analysis long 2023-03-01 2024-02-28 --model-path {tmp_path} --obs-path {tmp_path} --data-path {tmp_path} --coldata-path {tmp_path} --name 'Test' --medianscores"
    result = runner.invoke(app, options.split())
    assert result.exit_code == 0
    assert "Running CAMS2_83 Specific Statistics, cache is not cleared" in caplog.text
    assert "Failed to read model variable" in caplog.text


def test_config_options(
    tmp_path: Path,
    caplog,
):
    options = f"forecast week 2024-03-16 2024-03-23 --model-path {tmp_path} --obs-path {tmp_path} --data-path {tmp_path} --coldata-path {tmp_path} --id test_config --fairmode --addmap --addseasons"
    runner.invoke(app, options.split())
    assert "'add_model_maps': True," in caplog.text
    assert "'add_seasons': True," in caplog.text
    assert "'exp_id': 'test_config'," in caplog.text
    assert "'use_fairmode': True," in caplog.text
    assert "'periods': ['20240316-20240323']," in caplog.text


@pytest.mark.parametrize(
    "evaltype,use_fairmode_flag,use_cams2_83_fairmode_flag",
    [
        pytest.param(
            "week",
            True,
            False,
            id="week",
        ),
        pytest.param(
            "season",
            False,
            True,
            id="season",
        ),
        pytest.param(
            "long",
            False,
            True,
            id="long",
        ),
    ],
)
def test_make_config(
    tmp_path: Path,
    evaltype: str,
    use_fairmode_flag: bool,
    use_cams2_83_fairmode_flag: bool,
):
    start_date = date(2025, 3, 1)
    end_date = date(2025, 3, 8)
    leap = 2
    models = [ModelName("emep"), ModelName("chimere")]
    id = "test"
    name = "test"
    description = "test"
    eval_type = EvalType(evaltype)
    run_type = RunType("forecast")

    cfg = make_config(
        start_date=start_date,
        end_date=end_date,
        leap=leap,
        model_path=tmp_path,
        obs_path=tmp_path,
        data_path=tmp_path,
        coldata_path=tmp_path,
        id=id,
        name=name,
        description=description,
        eval_type=eval_type,
        run_type=run_type,
        models=models,
        add_map=True,
        only_map=True,
        add_seasons=True,
        fairmode=True,
        medianscores=True,
    )
    assert cfg["periods"] == ["20250301-20250308"]
    assert cfg["add_model_maps"]
    assert cfg["only_model_maps"]
    assert cfg["use_cams2_83_fairmode"] == use_cams2_83_fairmode_flag
    assert cfg["use_fairmode"] == use_fairmode_flag



import os

from pathlib import Path

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    field_validator,
)


from pyaerocom import const
from pyaerocom.aeroval.modelmaps_helpers import CONTOUR, OVERLAY

PLOT_TYPE_OPTIONS = ({OVERLAY}, {CONTOUR}, {OVERLAY, CONTOUR})


class OutputPaths(BaseModel):
    """
    Setup class for output paths of json files and co-located data

    This interface generates all paths required for an experiment.

    Attributes
    ----------
    proj_id : str
        project ID
    exp_id : str
        experiment ID
    json_basedir : str, Path
    avdb_resource : str, Path, None
        An aerovaldb resource identifier as expected by aerovaldb.open()[1].
        If not provided, pyaerocom will fall back to using json_basedir, for
        backwards compatibility.

        [1] https://aerovaldb.readthedocs.io/en/latest/api.html#aerovaldb.open
    """

    # Pydantic ConfigDict
    model_config = ConfigDict(arbitrary_types_allowed=True)

    _JSON_SUBDIRS: list[str] = [
        "map",
        "ts",
        "ts/diurnal",
        "scat",
        "hm",
        "hm/ts",
        "contour",
        "profiles",
        "overlay",
    ]
    avdb_resource: Path | str | None = None

    json_basedir: Path | str = Field(
        default=os.path.join(const.OUTPUTDIR, "aeroval/data"), validate_default=True
    )
    coldata_basedir: Path | str = Field(
        default=os.path.join(const.OUTPUTDIR, "aeroval/coldata"), validate_default=True
    )

    @field_validator("json_basedir", "coldata_basedir")
    @classmethod
    def validate_basedirs(cls, v):
        if not os.path.exists(v):
            tmp = Path(v) if isinstance(v, str) else v
            tmp.mkdir(parents=True, exist_ok=True)
        return v

    proj_id: str
    exp_id: str

    def _check_init_dir(self, loc, assert_exists):
        if assert_exists and not os.path.exists(loc):
            os.makedirs(loc)
        return loc

    def get_coldata_dir(self, assert_exists=True):
        loc = os.path.join(self.coldata_basedir, self.proj_id, self.exp_id)
        return self._check_init_dir(loc, assert_exists)

    def get_json_output_dirs(self, assert_exists=True):
        out = {}
        base = os.path.join(self.json_basedir, self.proj_id, self.exp_id)
        for subdir in self._JSON_SUBDIRS:
            loc = self._check_init_dir(os.path.join(base, subdir), assert_exists)
            out[subdir] = loc
        # for cams2_83 the extra 'forecast' folder will contain the median scores if computed
        if self.proj_id == "cams2-83":
            loc = self._check_init_dir(os.path.join(base, "forecast"), assert_exists)
            out["forecast"] = loc
        return out



"""
Module containing time resampling functionality
"""

import logging

import pandas as pd
import xarray as xarr

from pyaerocom.exceptions import TemporalResolutionError
from pyaerocom.helpers import isnumeric, resample_time_dataarray, resample_timeseries
from pyaerocom.units.datetime import TsType

logger = logging.getLogger(__name__)


class TimeResampler:
    """Object that can be use to resample timeseries data

    It supports hierarchical resampling of :class:`xarray.DataArray` objects
    and :class:`pandas.Series` objects.

    Hierarchical means, that resampling constraints can be applied for each
    level, that is, if hourly data is to be resampled to monthly, it may be
    specified to first required minimum number of hours per day, and minimum
    days per month, to create the output data.
    """

    AGGRS_UNIT_PRESERVE = ("mean", "median", "std", "max", "min")
    DEFAULT_HOW = "mean"

    def __init__(self, input_data=None):
        self.last_setup = None
        # the following attribute is updated whenever a resampling operation is
        # performed and it will check if any of the specified resampling
        # aggregators invalidates unit preservation (e.g. using how=add for
        # for accumulating precipitation...). See also attr. AGGRS_UNIT_PRESERVE
        self._last_units_preserved = None
        self._input_data = None

        if input_data is not None:
            self.input_data = input_data

    @property
    def last_units_preserved(self):
        """Boolean indicating if last resampling operation preserves units"""
        if self._last_units_preserved is None:
            raise AttributeError("Please call resample first...")
        return self._last_units_preserved

    @property
    def input_data(self):
        """Input data object that is to be resampled"""
        return self._input_data

    @input_data.setter
    def input_data(self, val):
        if not isinstance(val, pd.Series | xarr.DataArray):
            raise ValueError("Invalid input: need Series or DataArray")
        self._input_data = val

    @property
    def fun(self):
        """Resampling method (depends on input data type)"""
        if isinstance(self.input_data, pd.Series):
            return resample_timeseries
        return resample_time_dataarray

    def _get_resample_how(self, fr, to, how):
        if not isinstance(how, str | dict):
            val = self.DEFAULT_HOW
        elif isinstance(how, dict):
            if to.val in how and fr.val in how[to.val]:
                val = how[to.val][fr.val]
            else:
                val = self.DEFAULT_HOW
        else:
            val = how
        return val

    def _get_idx_entry(self, fr, to, min_num_obs, how):
        min_num = fr.get_min_num_obs(to, min_num_obs)

        _how = self._get_resample_how(fr, to, how)

        return (to.val, min_num, _how)

    def _gen_idx(self, from_ts_type, to_ts_type, min_num_obs, how):
        """Generate hierarchical resampling index

        Return
        ------
        list
            list (can be considered the iterator) of 3-element tuples for each\
            resampling step, containing

            - frequency to which the current is converted
            - minimum number of not-NaN values required for that step
            - aggregator to be used (e.g. mean, median, ...)

        """
        if isnumeric(min_num_obs):
            if not isinstance(how, str):
                raise ValueError(
                    f"Error initialising resampling constraints. "
                    f"min_num_obs is numeric ({min_num_obs}) and input how is {how} "
                    f"(would need to be string, e.g. mean)"
                )
            return [(to_ts_type.val, int(min_num_obs), how)]
        if not isinstance(min_num_obs, dict):
            raise ValueError(
                f"Invalid input for min_num_obs, need dictionary or integer, got {min_num_obs}"
            )

        base_freqs = TsType.VALID

        start_base = base_freqs.index(from_ts_type.base)
        stop_base = base_freqs.index(to_ts_type.base)

        last_from = from_ts_type
        idx = []
        # loop from next base freq to end base freq, note that min_num_obs as
        # well as input freqs may have multiplication factors, which may
        # require min_num_obs values to be updated accordingly
        for i in range(start_base + 1, stop_base + 1):
            to_base = TsType(base_freqs[i])
            try:
                entry = self._get_idx_entry(last_from, to_base, min_num_obs, how)
                idx.append(entry)
                last_from = to_base
            except (TemporalResolutionError, ValueError):
                continue
        if len(idx) == 0 or not idx[-1][0] == to_ts_type.val:
            try:
                last_entry = self._get_idx_entry(last_from, to_ts_type, min_num_obs, how)
            except (TemporalResolutionError, ValueError):
                _how = self._get_resample_how(last_from, to_ts_type, how)
                last_entry = (to_ts_type.val, 0, _how)
            idx.append(last_entry)
        return idx

    def resample(
        self, to_ts_type, input_data=None, from_ts_type=None, how=None, min_num_obs=None, **kwargs
    ):
        """Resample input data

        Parameters
        ----------
        to_ts_type : str or TsType
            output resolution
        input_data : pandas.Series or xarray.DataArray
            data to be resampled
        from_ts_type : str or TsType, optional
            current temporal resolution of data
        how : str
            string specifying how the data is to be aggregated, default is mean
        min_num_obs : dict or int, optional
            integer or nested dictionary specifying minimum number of
            observations required to resample from higher to lower frequency.
            For instance, if `input_data` is hourly and `to_ts_type` is
            monthly, you may specify something like::

                min_num_obs =
                    {'monthly'  :   {'daily'  : 7},
                     'daily'    :   {'hourly' : 6}}

            to require at least 6 hours per day and 7 days per month.

        **kwargs
           additional input arguments passed to resampling method

        Returns
        -------
        pandas.Series or xarray.DataArray
            resampled data object
        """
        if how is None:
            how = "mean"

        if how in self.AGGRS_UNIT_PRESERVE:
            self._last_units_preserved = True
        else:
            self._last_units_preserved = False

        if not isinstance(to_ts_type, TsType):
            to_ts_type = TsType(to_ts_type)

        if str(from_ts_type) == "native":
            from_ts_type = None

        if from_ts_type is None:
            if min_num_obs is not None:
                logger.warning("setting min_num_obs to None since from_ts_type is not specified")
                min_num_obs = None
        elif isinstance(from_ts_type, str):
            from_ts_type = TsType(from_ts_type)

        if input_data is not None:
            self.input_data = input_data
        if self.input_data is None:
            raise ValueError("Please provide data (Series or DataArray)")

        self.last_setup = dict(min_num_obs=min_num_obs, how=how)

        if from_ts_type is None:  # native == unknown
            freq = to_ts_type.to_pandas_freq()
            data_out = self.fun(self.input_data, freq=freq, how=how, **kwargs)
        elif to_ts_type > from_ts_type:
            raise TemporalResolutionError(
                f"Cannot resample time-series from {from_ts_type} to {to_ts_type}"
            )
        elif to_ts_type == from_ts_type:
            logger.debug(
                f"Input time frequency {to_ts_type.val} equals current frequency of data. "
                f"Resampling will be applied anyways which will introduce NaN values "
                f"at missing time stamps"
            )

            freq = to_ts_type.to_pandas_freq()
            self._last_units_preserved = True
            data_out = self.fun(self.input_data, freq=freq, how="mean", **kwargs)

        elif min_num_obs is None:
            freq = to_ts_type.to_pandas_freq()
            if not isinstance(how, str):
                raise ValueError(
                    f"Temporal resampling without constraints can only use string type "
                    f"argument how (e.g. how=mean). Got {how}"
                )

            data_out = self.fun(self.input_data, freq=freq, how=how, **kwargs)
        else:
            _idx = self._gen_idx(from_ts_type, to_ts_type, min_num_obs, how)
            data_out = self.input_data
            aggrs = []
            for to_ts_type, mno, rshow in _idx:
                freq = TsType(to_ts_type).to_pandas_freq()
                data_out = self.fun(data_out, freq=freq, how=rshow, min_num_obs=mno, **kwargs)
                aggrs.append(rshow)

            if all([x in self.AGGRS_UNIT_PRESERVE for x in aggrs]):
                self._last_units_preserved = True
            else:
                self._last_units_preserved = False
        return data_out



from __future__ import annotations

from pathlib import Path

import pytest
import numpy as np

from pyaerocom.io.read_eprofile import ReadEprofile
from pyaerocom import const, VerticalProfile, UngriddedData


ROOT: str = const.OBSLOCS_UNGRIDDED["Eprofile-test"]

TEST_FILES: list[str] = [
    f"{ROOT}/AP_0-20000-0-01001-A-2025-01-01.nc",
    f"{ROOT}/AP_0-20000-0-06380-A-2025-01-01.nc",
]


def test_all_files_exist():
    for file in TEST_FILES:
        assert Path(file).exists()


@pytest.mark.parametrize(
    "num,vars_to_retrieve",
    [
        (0, "bsc1064aer"),
    ],
)
def test_ReadEprofile_read_file(num: int, vars_to_retrieve: list[str]):
    read = ReadEprofile()
    read.files = paths = TEST_FILES
    stat = read.read_file(paths[num], vars_to_retrieve)

    assert "data_level" in stat
    assert "wavelength_emis" in stat
    assert "station_coords" in stat

    assert vars_to_retrieve in stat.var_info
    assert stat.var_info[vars_to_retrieve]["unit_ok"]
    assert not stat.var_info[vars_to_retrieve]["err_read"]
    assert stat.var_info[vars_to_retrieve]["outliers_removed"]

    bsc1064aer = stat.bsc1064aer
    assert isinstance(bsc1064aer, VerticalProfile)
    assert len(bsc1064aer.data) > 0
    assert isinstance(np.min(bsc1064aer.altitude), float)
    assert isinstance(np.max(bsc1064aer.altitude), float)


def test_ReadEprofile_read_file_error():
    read = ReadEprofile()
    read.files = paths = TEST_FILES
    with pytest.raises(ValueError) as e:
        read.read_file(paths[0], "invalidvar")
    assert str(e.value).endswith("is not supported")


def test_ReadEprofile_read():
    read = ReadEprofile()
    read.files = TEST_FILES
    data = read.read(vars_to_retrieve="bsc1064aer")
    assert isinstance(data, UngriddedData)
    assert len(data.metadata) == len(TEST_FILES)
    assert len(data.shape) == 2



"""
Helper methods for computation of trends

Note
----
Most methods here are private and not to be used directly. Please use
:class:`TrendsEngine` instead.
"""

import numpy as np
import pandas as pd

SEASONS = {"spring": [3, 4, 5], "summer": [6, 7, 8], "autumn": [9, 10, 11], "winter": [12, 1, 2]}

SEASON_CODES = {"spring": "MAM", "summer": "JJA", "autumn": "SON", "winter": "DJF"}

MONTHS_CODES = {
    "MAM": "spring",
    "JJA": "summer",
    "SON": "autumn",
    "DJF": "winter",
    "all": "all",
}


def _init_trends_result_dict(start_yr):
    keys = [
        "pval",
        "m",
        "m_err",
        "n",
        "y_mean",
        "y_min",
        "y_max",
        "slp",
        "slp_err",
        "reg0",  # data specific
        f"slp_{start_yr}",  # period specific
        f"slp_{start_yr}_err",  # period specific
        f"reg0_{start_yr}",  # period specific
        "data",
    ]
    return dict.fromkeys(keys)


def _compute_trend_error(m, m_err, v0, v0_err):
    """Computes error of trend estimate using gaussian error propagation

    The (normalised) trend is computed as T = m / v0

    where m denotes the slope of a regression line and v0 denotes the
    normalistation value. This method computes the uncertainty of T (delta_T)
    using gaussian error propagation of uncertainties accompanying m and v0.

    Parameters
    ----------
    m : float
        slope in units of <U> yr-1 (where <U> denotes the unit of the data).
        (m -> "montant").
    m_err : float
        slope error (same unit as `m`)
    v0 : float
        normalisation value in units of <U>
    v0_err : float
        error of `v0` (same units as `v0`)

    Returns
    -------
    float
        error of T in computed using gaussian error propagation of trend
        formula in units of %/yr
    """

    delta_sl = m_err / v0
    delta_ref = m * v0_err / v0**2
    return np.sqrt(delta_sl**2 + delta_ref**2) * 100


def _get_season(mon):
    for seas, months in SEASONS.items():
        if mon in months:
            return seas


def _get_unique_seasons(idx):
    seasons = []
    mons = idx.month
    for mon in mons:
        seas = _get_season(mon)
        if seas not in seasons:
            seasons.append(seas)
    return seasons


def _get_season_from_months(months: str) -> str:
    if months not in MONTHS_CODES:
        raise ValueError(f"{months} is not a valid season")
    return MONTHS_CODES[months]


def _mid_season(seas, yr):
    if seas == "spring":
        return np.datetime64(f"{yr}-04-15")
    if seas == "summer":
        return np.datetime64(f"{yr}-07-15")
    if seas == "autumn":
        return np.datetime64(f"{yr}-10-15")
    if seas == "winter":
        return np.datetime64(f"{yr}-01-15")
    if seas == "all":
        return np.datetime64(f"{yr}-06-15")
    raise ValueError("Invalid input for season (seas):", seas)


def _start_season(seas, yr):
    if seas == "spring":
        return f"{yr}-03-01"
    if seas == "summer":
        return f"{yr}-06-01"
    if seas == "autumn":
        return f"{yr}-09-01"
    if seas == "winter":
        return f"{yr - 1}-12-01"
    if seas == "all":
        return f"{yr}-01-01"
    raise ValueError("Invalid input for season (seas):", seas)


def _end_season(seas, yr):
    if seas == "spring":
        return f"{yr}-06-01"
    if seas == "summer":
        return f"{yr}-09-01"
    if seas == "autumn":
        return f"{yr}-12-01"
    if seas == "winter":
        return f"{yr}-03-01"
    if seas == "all":
        return f"{yr}-01-01"
    raise ValueError("Invalid input for season (seas):", seas)


def _find_area(lat, lon, regions_dict=None):
    """Find area corresponding to input lat/lon coordinate

    Parameters
    ----------
    lat : float
        latitude
    lon : float
        longitude

    Returns
    -------
    str
        name of region
    """
    from pyaerocom.region import find_closest_region_coord

    reg = find_closest_region_coord(lat, lon, regions_how="default")[0]
    if regions_dict is not None and reg in regions_dict:
        return regions_dict[reg]
    return reg


def _years_from_periodstr(period):
    """Convert period str to start / stop years

    Parameters
    ----------
    period : str
        period str, e.g. '1990-2010'

    Returns
    -------
    int
        start year
    int
        stop year
    """
    return [int(x) for x in period.split("-")]


def _start_stop_period(period):
    """Convert period str to start / stop dates

    Parameters
    ----------
    period : str
        period str, e.g. '1990-2010'

    Returns
    -------
    date
        start datetime
    date
        stop datetime
    """
    from datetime import date

    (
        y0,
        y1,
    ) = _years_from_periodstr(period)
    return (date(y0, 1, 1), date(y1, 12, 31))


def _seas_slice(yr, season):
    pass


def _get_yearly(data, seas, start_yr):
    dates = []
    values = []
    yrs = np.unique(data.index.year)
    for yr in yrs:
        if yr < start_yr:  # winter
            continue

        if seas == "all":  # yearly trends
            subset = data.loc[str(yr)]

        else:
            start = _start_season(seas, yr)
            stop = _end_season(seas, yr)
            subset = data.loc[start:stop]

        if len(subset) == 0:
            val = np.nan
        elif np.isnan(subset.values).all():
            val = np.nan

        elif seas == "all":
            d = subset.index
            valid_mask = ~np.isnan(subset.values)
            # seasons = _get_unique_seasons(d) # original (EMEP raport)
            seasons = _get_unique_seasons(d[valid_mask])

            if len(seasons) == 4:
                val = np.nanmean(subset.values)
            else:
                val = np.nan
        else:
            val = np.nanmean(subset.values)

        dates.append(_mid_season(seas, yr))
        values.append(val)

    return pd.Series(values, index=dates)


def _init_period_dates(start_year, stop_year, season):
    start_date = _mid_season(season, start_year)
    stop_date = _mid_season(season, stop_year)

    # datetime index covering whole input period (data may not be fully covered
    # in the whole period)
    period_index = pd.date_range(start=start_date, end=stop_date, freq=pd.DateOffset(years=1))

    num_dates_period = period_index.values.astype("datetime64[Y]").astype(np.float64)
    return (start_date, stop_date, period_index, num_dates_period)



from __future__ import annotations

import logging

from pyaro import list_timeseries_engines, open_timeseries
from pyaro.timeseries.Wrappers import VariableNameChangingReader

from pyaerocom.io.pyaro.postprocess import PostProcessingReader
from pyaerocom.io.pyaro.pyaro_config import PyaroConfig
from pyaerocom.io.readungriddedbase import ReadUngriddedBase
from pyaerocom.ungriddeddata_structured import UngriddedDataStructured
from pyaro.timeseries import Reader

logger = logging.getLogger(__name__)


class ReadPyaro(ReadUngriddedBase):
    __version__ = "1.2.1"

    SUPPORTED_DATASETS = list(list_timeseries_engines().keys())

    def __init__(self, config: PyaroConfig) -> None:
        self.config: PyaroConfig = config

        self._check_id()

        self.converter = PyaroToUngriddedData(self.config)
        self.reader = self.converter.reader
        self._data_dir = self.config.filename_or_obj_or_url
        self._data_name = self.config.name
        self._data_id = self.config.name

    """
    Definition of abstract methods from ReadUngriddedBase
    """

    @property
    def DATA_ID(self):
        return self._data_name

    @property
    def PROVIDES_VARIABLES(self):
        """
        return self.reader.get_variables()
        """
        return self.reader.variables()

    @property
    def DEFAULT_VARS(self):
        return self.PROVIDES_VARIABLES

    @property
    def TS_TYPE(self):
        """
        To be provided by the reader or engine
        """
        return "undefined"

    @property
    def _FILEMASK(self):
        return self.config.filename_or_obj_or_url

    @staticmethod
    def get_pyaro_readers():
        return list_timeseries_engines()

    def read(self, vars_to_retrieve=None, files=..., first_file=None, last_file=None):
        return self.converter.read(vars_to_retrieve=vars_to_retrieve)

    def read_file(self, filename, vars_to_retrieve=None):
        return self.converter.read(vars_to_retrieve)

    def _check_id(self):
        avail_readers = list_timeseries_engines()
        if self.config.reader_id not in avail_readers:
            logger.warning(
                f"Could not find {self.config.reader_id} in list of available Pyaro readers: {avail_readers}"
            )


class PyaroToUngriddedData:
    def __init__(self, config: PyaroConfig) -> None:
        self.data: UngriddedDataStructured = UngriddedDataStructured()
        self.config = config
        self.reader: Reader = self._open_reader()

    def _open_reader(self) -> Reader:
        reader_id = self.config.reader_id
        if self.config.model_extra is not None:
            kwargs = self.config.model_extra
        else:
            kwargs = {}

        reader = open_timeseries(
            reader_id,
            self.config.filename_or_obj_or_url,
            filters=self.config.filters,
            **kwargs,
        )
        if self.config.name_map is not None:
            reader = VariableNameChangingReader(
                reader,
                self.config.name_map,
            )
        if self.config.post_processing is not None:
            reader = PostProcessingReader(
                reader,
                self.config.post_processing,
            )
        return reader

    def get_variables(self) -> list[str]:
        return self.reader.variables()

    def read(self, vars_to_retrieve=None) -> UngriddedDataStructured:
        allowed_vars = self.get_variables()
        if vars_to_retrieve is None:
            vars_to_retrieve = allowed_vars
        else:
            if isinstance(vars_to_retrieve, str):
                vars_to_retrieve = [vars_to_retrieve]

        vars_to_retrieve_clean = []
        for var in vars_to_retrieve:
            if var not in allowed_vars:
                logger.warning(
                    f"Variable {var} not in list over allowed variables for {self.config.reader_id}: {allowed_vars}"
                )
                continue
            vars_to_retrieve_clean.append(var)

        return UngriddedDataStructured.from_pyaro(
            self.config.name, self.reader, vars_to_retrieve_clean
        )



import pytest

from pyaerocom.units import Unit
from pyaerocom.units.units import UnitConversionCallbackInfo


@pytest.mark.parametrize(
    "unit,aerocom_var,to_unit,exp_mul",
    (
        ("meter", None, "km", 0.001),
        ("km", None, "meter", 1000),
        ("ug S/m3", "concso2", "ug m-3", 1.997935),
    ),
)
def test_PyaerocomUnit_custom_scaling(
    unit: str, aerocom_var: str | None, to_unit: str, exp_mul: float
):
    u1 = Unit(unit, aerocom_var=aerocom_var)
    u2 = Unit(to_unit, aerocom_var=aerocom_var)

    assert u1.convert(1, other=u2) == pytest.approx(exp_mul)
    assert u1.convert(1, other=u2, inplace=True) == pytest.approx(exp_mul)


@pytest.mark.parametrize(
    "unit,tstype,output_cf_unit",
    (
        ("mg m-2", "daily", "mg m-2 d-1"),
        ("mg m-2", "yearly", "mg m-2 yr-1"),
        ("mg m-2 d-1", "daily", "mg m-2 d-1"),
    ),
)
def test_PyaerocomUnit_implicit_frequency(unit: str, tstype: str | None, output_cf_unit: str):
    u = Unit(unit, aerocom_var="depdust", ts_type=tstype)
    assert str(u) == output_cf_unit


def test_PyaerocomUnit_conversion_callback():
    u = Unit("mg m-2", aerocom_var="depdust", ts_type="daily")

    callback_ran = False

    def callback(info: UnitConversionCallbackInfo):
        nonlocal callback_ran
        callback_ran = True

        assert info.factor == pytest.approx(1 / 24)
        assert info.from_aerocom_var == "depdust"
        assert info.from_ts_type == "daily"
        assert info.from_cf_unit == "mg m-2 d-1"
        assert info.to_cf_unit == "mg m-2 h-1"

    u.convert(1, "mg m-2 h-1", callback=callback)

    assert callback_ran


def test_origin():
    assert Unit("ug S/m3").origin == "ug S/m3"


@pytest.mark.parametrize(
    "from_unit,to_unit,is_convertible",
    (
        ("m", "km", True),
        ("m", "kg", False),
        (Unit("mg S"), Unit("mg"), False),
        (Unit("mg S", species="SO4"), Unit("mg", species="SO4"), True),
        (Unit("mg S", aerocom_var="concso2"), Unit("mg", aerocom_var="concso2"), True),
        (Unit("mg", species="SO4"), Unit("mg S", species="SO4"), True),
        (Unit("mg", aerocom_var="concso2"), Unit("mg S", aerocom_var="concso2"), True),
        (Unit("mg S", species="SO2"), Unit("mg N", species="SO2"), False),
    ),
)
def test_is_convertible(from_unit: str | Unit, to_unit: str | Unit, is_convertible: bool):
    if isinstance(from_unit, str):
        from_unit = Unit(from_unit)
    if isinstance(to_unit, str):
        to_unit = Unit(to_unit)

    assert from_unit.is_convertible(to_unit) == is_convertible


def test_is_dimensionless():
    assert not Unit("m").is_dimensionless()
    assert Unit("1").is_dimensionless()


@pytest.mark.parametrize(
    "from_unit,to_unit,is_equal,aerocom_var,ts_type",
    (
        ("meter", "meter", True, None, None),
        ("meter", "cm", False, None, None),
        ("ug S/m3", "1.9979354436301264 ug m-3", True, "concso2", None),
        ("g m-2", "g m-2 d-1", True, "depdust", "daily"),
    ),
)
def test_equality(
    from_unit: str, to_unit: str, is_equal: bool, aerocom_var: str | None, ts_type: str | None
):
    assert (Unit(from_unit, aerocom_var=aerocom_var, ts_type=ts_type) == Unit(to_unit)) == is_equal
    assert (Unit(from_unit, aerocom_var=aerocom_var, ts_type=ts_type) != Unit(to_unit)) != is_equal


@pytest.mark.parametrize(
    "unit,nominator,denominator",
    (("mg", "mg", ""), ("mg N / m2 d", "mg N", "m2 d"), ("mg N m-2 d-1", "mg N", "m-2 d-1")),
)
def test_nominator_denominator(unit: str, nominator: str, denominator: str):
    u = Unit(unit)
    assert u._origin_nominator == nominator
    assert u._origin_denominator == denominator


@pytest.mark.parametrize(
    "unit,species,var,exp_element,exp_species",
    (
        ("mg", None, None, None, None),
        ("mg N", None, None, None, None),
        ("mg N", "NH3", None, "N", "NH3"),
        ("mg N", None, "drynh3", "N", "NH3"),
    ),
)
def test_species_and_element_detection(
    unit: str,
    species: str | None,
    var: str | None,
    exp_element: str | None,
    exp_species: str | None,
):
    u = Unit(unit, species=species, aerocom_var=var)
    assert u._element == exp_element
    assert u._species == exp_species


@pytest.mark.parametrize(
    "from_unit,to_unit,species,conversion_fac",
    (
        (
            "kg N",
            "kg",
            "NH3",
            17.03052 / 14.0067,  # Molecular mass ratio of NH3/N
        ),
        ("mg N", "kg", "NH3", (17.03052 / 14.0067) / 1_000_000),
        ("mg N s-1", "kg s-1", "NH3", (17.03052 / 14.0067) / 1_000_000),
        ("mg N s-1", "kg h-1", "NH3", 3_600 * (17.03052 / 14.0067) / (1_000_000)),
    ),
)
def test_unit_conversion(from_unit: str, to_unit: str, species: str, conversion_fac: float):
    u = Unit(from_unit, species=species)
    fac = u.convert(1, to_unit, species=species)
    assert fac == pytest.approx(conversion_fac)


@pytest.mark.parametrize(
    "from_unit,to_unit",
    (
        (  # No known reference species to do conversion.
            Unit("kg S"),
            Unit("kg"),
        ),
        (  # Element don't match
            Unit(
                "kg N",
            ),
            Unit("kg S"),
        ),
    ),
)
def test_unit_conversion_fails(from_unit: Unit, to_unit: Unit):
    with pytest.raises(Exception):
        from_unit.convert(1, to_unit)



import pytest

from pyaerocom.helpers import make_dummy_cube_latlon


def test_convert_units():
    cube = make_dummy_cube_latlon()
    assert (cube.data == 1).all()

    cube.units = "kg h-1"
    cube.convert_units("kg s-1")
    assert cube.data.mean() == pytest.approx(1 / 3600)


# =============================================================================
#
# cube = pya.helpers.make_dummy_cube_latlon()
# cube.var_name = 'wetso4'
# cube.units = 'mg S m-2 h-1'
#
# data = pya.GriddedData(cube)
#
# print(data)
# =============================================================================



import numpy as np
import pandas as pd
import pytest
import xarray as xr

from pyaerocom import StationData, helpers
from pyaerocom.exceptions import DataCoverageError, TemporalResolutionError
from pyaerocom.units import UnitConversionError


def test_get_standarad_name():
    assert (
        helpers.get_standard_name("od550aer")
        == "atmosphere_optical_thickness_due_to_ambient_aerosol_particles"
    )


@pytest.mark.parametrize("val", [3, 3.3455, complex(1, 2)])
def test_isnumeric(val):
    assert helpers.isnumeric(val)


@pytest.mark.parametrize(
    "val,result",
    [
        ((0, 1), True),
        ([10, 20], True),
        ([10, 20, 30], False),
    ],
)
def test_isrange(val, result):
    assert helpers.isrange(val) == result


@pytest.mark.parametrize(
    "use,var_name,pref_attr,sort_by_largest,fill_missing_nan,add_meta_keys,num,tst,mean",
    [
        ("concpm10", "concpm10", None, True, True, None, 730, "daily", 17.93),
        ("concpm10", "concpm10", "awesomeness", True, True, None, 730, "daily", 17.93),
        ("concpm10", "concpm10", "awesomeness", False, True, None, 730, "daily", 15),
        ("od550aer", "od550aer", None, True, True, None, 67, "60daily", 0.51),
        ("od550aer", "od550aer", "awesomeness", True, True, None, 67, "60daily", 0.59),
        ("od550aer", "od550aer", "awesomeness", False, True, None, 67, "60daily", 0.51),
    ],
)
def test_merge_station_data(
    statlist,
    use,
    var_name,
    pref_attr,
    sort_by_largest,
    fill_missing_nan,
    add_meta_keys,
    num,
    tst,
    mean,
):
    stats = [x.copy() for x in statlist[use]]
    stat = helpers.merge_station_data(
        stats, var_name, pref_attr, sort_by_largest, fill_missing_nan, add_meta_keys
    )
    assert isinstance(stat, StationData)
    vardata = stat[var_name]
    assert len(vardata) == num
    assert stat.get_var_ts_type(var_name) == tst
    assert np.mean(vardata) == pytest.approx(mean, rel=1e-2)


@pytest.mark.parametrize(
    "use,exception,error",
    [
        ("concpm10_X2", UnitConversionError, "failed to convert unit from mole mole-1 to ug m-3"),
        ("concpm10_X", TemporalResolutionError, "Invalid input for ts_type daily"),
        ("od550aer", DataCoverageError, "All input stations must contain concpm10 data"),
    ],
)
def test_merge_station_data_error(statlist, use, exception, error):
    stats = [x.copy() for x in statlist[use]]
    with pytest.raises(exception) as e:
        helpers.merge_station_data(stats, "concpm10")
    assert str(e.value).startswith(error)


def test__get_pandas_freq_and_offset():
    val = helpers._get_pandas_freq_and_offset("monthly")
    assert val == ("MS", pd.Timedelta(14, "d"))


@pytest.fixture(scope="module")
def fake_hourly_ts():
    time = pd.date_range("2018-01-10T00:00:00", "2018-01-17T23:59:00", freq="h")
    xrange_modulation = np.linspace(0, np.pi * 40, len(time))
    signal = np.sin(xrange_modulation)
    return pd.Series(signal, time)


@pytest.mark.parametrize(
    "freq,how,min_num_obs,num,avg",
    [
        ("daily", "mean", 10000, 8, np.nan),
        ("daily", "50percentile", 10000, 8, np.nan),
        ("daily", "mean", 23, 8, 0),
        ("daily", "50percentile", 23, 8, 0),
        ("yearly", "50percentile", 23, 1, 0),
        ("yearly", "50percentile", 192, 1, 0),
        ("daily", "50percentile", None, 8, 0),
        ("yearly", "50percentile", 193, 1, np.nan),
        ("daily", "median", None, 8, 0.0),
        ("monthly", "mean", None, 1, 0),
        ("daily", "mean", None, 8, 0),
        ("daily", "1percentile", None, 8, -1),
        ("daily", "25percentile", None, 8, -0.64),
        ("daily", "75percentile", None, 8, 0.64),
    ],
)
@pytest.mark.filterwarnings("ignore:Mean of empty slice:RuntimeWarning")
def test_resample_timeseries(fake_hourly_ts, freq, how, min_num_obs, num, avg):
    s1 = helpers.resample_timeseries(fake_hourly_ts, freq=freq, how=how, min_num_obs=min_num_obs)
    assert len(s1) == num
    assert np.nanmean(s1) == pytest.approx(avg, abs=1e-2, nan_ok=True)


def test_same_meta_dict():
    d1 = dict(
        station_name="bla", station_id="blub", latitude=33, longitude=15, altitude=400, PI="pi1"
    )
    d2 = dict(
        station_name="bla", station_id="blub1", latitude=33, longitude=15, altitude=401, PI="pi2"
    )

    assert helpers.same_meta_dict(d1, d2) is False


def test_to_pandas_timestamp():
    pass


def test_to_datetime64():
    pass


def test_is_year():
    assert helpers.is_year(2010)


def test_start_stop():
    pass


def test_datetime2str():
    pass


def test_start_stop_str():
    pass


def test_start_stop_from_year():
    start, stop = helpers.start_stop_from_year(2000)
    assert start == pd.Timestamp("2000")
    assert stop == pd.Timestamp("2000-12-31 23:59:59")


@pytest.mark.parametrize(
    "input,expected",
    [("20100101", "20100101"), (helpers.to_pandas_timestamp("20100101"), "20100101")],
)
def test_to_datestring_YYYYMMDD(input, expected):
    assert helpers.to_datestring_YYYYMMDD(input) == expected


def get_constraint():
    pass


def test_get_lat_rng_constraint():
    pass


def test_get_lon_rng_constraint():
    pass


def test_get_time_rng_constraint():
    pass


def test_extract_latlon_dataarray():
    cube = helpers.make_dummy_cube_latlon(
        lat_res_deg=1, lon_res_deg=1, lat_range=[10, 20], lon_range=[10, 20]
    )
    data = xr.DataArray.from_iris(cube)
    # First coordinate does not exist in the dataarray.
    lat = [15, 15, 18]
    lon = [1, 15, 18]
    subset = helpers.extract_latlon_dataarray(data, lat, lon, check_domain=True)
    assert isinstance(subset, xr.DataArray)
    assert len(subset.lat) == len(lat) - 1
    assert len(subset.lon) == len(lon) - 1


def test_extract_latlon_dataarray_no_matches():
    cube = helpers.make_dummy_cube_latlon(
        lat_res_deg=1, lon_res_deg=1, lat_range=[10, 20], lon_range=[10, 20]
    )
    data = xr.DataArray.from_iris(cube)
    helpers.extract_latlon_dataarray(data, [15], [15], check_domain=True)


@pytest.mark.parametrize(
    "lat,lon",
    [
        ([], []),
        ([1, 2], [-1, 2]),
    ],
)
def test_extract_latlon_dataarray_no_matches_error(lat, lon):
    cube = helpers.make_dummy_cube_latlon(
        lat_res_deg=1, lon_res_deg=1, lat_range=[10, 20], lon_range=[10, 20]
    )
    data = xr.DataArray.from_iris(cube)
    with pytest.raises(DataCoverageError) as e:
        helpers.extract_latlon_dataarray(data, lat, lon, check_domain=True)
    assert str(e.value) == "Coordinates not found in dataarray"


def test_make_dummy_cube():
    # make a dummy cube of an arbitrary variable name over one year
    cube = helpers.make_dummy_cube("concpm10", start_yr=2020, stop_yr=2021)
    data = xr.DataArray.from_iris(cube)
    # First coordinate does not exist in the dataarray.
    lat = [90, -67.5, 22.2]
    lon = [180, -135.0, 45.0]
    subset = helpers.extract_latlon_dataarray(data, lat, lon, check_domain=True)
    assert isinstance(subset, xr.DataArray)
    assert len(subset.lat) == len(lat) - 1
    assert len(subset.lon) == len(lon) - 1



import logging

# from pathlib import Path
import numpy as np
import xarray as xr

from pyaerocom import ColocatedData
from pyaerocom.aeroval.experiment_output import ExperimentOutput
from pyaerocom.units.datetime import TsType

logger = logging.getLogger(__name__)


SPECIES = dict(
    concno2=dict(
        UrRV=0.24, RV=200, alpha=0.2, freq=TsType("hourly"), percentile=99.8, Np=5.2, Nnp=5.5
    ),
    conco3mda8=dict(
        UrRV=0.18, RV=120, alpha=0.79, freq=TsType("daily"), percentile=92.9, Np=11.0, Nnp=3.0
    ),
    concpm10=dict(
        UrRV=0.28, RV=50, alpha=0.25, freq=TsType("daily"), percentile=90.1, Np=20.0, Nnp=1.5
    ),
    concpm25=dict(
        UrRV=0.36, RV=25, alpha=0.5, freq=TsType("daily"), percentile=90.1, Np=20.0, Nnp=1.5
    ),
)

EXC_THRESHOLDS = dict(  # we assume all the units are ug/m3
    concpm25=25.0,
    concpm10=50.0,
    conco3mda8=120.0,
    concno2=200.0,
)


class FairmodeStatistics:
    """
    Class for computing the FAIRMODE statistics

    FAIRMODE is the Forum for Air Quality Modeling, an initiative to bring together air quality modelers and users.
        - Promote and Support the use of models by EU Member States
        - Emphasis is on model application for air quality policy (monitoring, regulation, etc.)
        - Develop harmonized set of tools to test whether or a not a model is fit for a given purpose
        - CAMS has to make use of FAIRMODE diagrams

    This module contains methods to compute the relevant FAIRMODE statistics.
    """

    species = SPECIES

    def __init__(self):
        pass

    def save_fairmode_stats(
        self,
        exp_output: ExperimentOutput,
        fairmode_stats: dict,
        obs_name: str,
        var_name_web: str,
        vert_code: str,
        modelname: str,
        model_var: str,
    ):
        for regname in fairmode_stats:
            exp_output.add_fairmode_entry(
                fairmode_stats[regname],
                regname,
                obs_name,
                var_name_web,
                vert_code,
                modelname,
                model_var,
            )

    def fairmode_statistics(self, coldata: ColocatedData, var_name: str):
        return self._get_stats(coldata.data, var_name, False)

    def _get_stats(
        self, data: xr.DataArray, var_name: str, use_weights: bool
    ) -> dict[str, dict[str, float]]:
        stations = data.station_name.values
        station_types = data.station_type.values

        obsvals = data.data[0]
        modvals = data.data[1]

        mask = np.isfinite(obsvals) & np.isfinite(modvals)

        obsmean = np.mean(obsvals, axis=0, where=mask)
        modmean = np.mean(modvals, axis=0, where=mask)

        obsstd = np.std(obsvals, axis=0, where=mask)
        modstd = np.std(modvals, axis=0, where=mask)

        diff = modvals - obsvals
        diffsquare = diff**2

        rms = np.sqrt(np.nanmean(diffsquare, axis=0, where=mask))
        bias = np.nanmean(diff, axis=0, where=mask)

        βRMSUt = self._βRMSU_t(obsvals, beta=1, var_name=var_name, mask=mask)
        βRMSUs = self._βRMSU_s(obsmean, beta=1, var_name=var_name)

        NMB = self._NMB(modvals, obsvals)
        R = self.pearson_R(obsvals, modvals)
        sign = self._fairmode_sign(modstd, obsstd, R)
        crms = self._crms(modstd, obsstd, R)
        mqi = self._mqi(rms, βRMSUt, beta=1)
        mb = self._mb(bias, βRMSUt, beta=1)
        beta_Hperc = self._beta_Hperc(obsvals, modvals, var_name)
        exceedances = self._exceedances(data=data, var_name=var_name)

        MPI_bias_t = self._MPI_bias_t(obsmean, modmean, βRMSUt)
        MPI_R_t = self._MPI_R_t(obsstd, modstd, R, βRMSUt)
        MPI_std_t = self._MPI_std_t(obsstd, modstd, βRMSUt)

        MPI_R_s = self._MPI_R_s(obsmean, modmean, βRMSUs)
        MPI_std_s = self._MPI_std_s(obsmean, modmean, βRMSUs)

        assert len(βRMSUt) == len(stations)
        assert len(sign) == len(stations)
        assert len(crms) == len(stations)
        assert len(bias) == len(stations)
        assert len(rms) == len(stations)
        assert len(mqi) == len(stations)
        assert len(mb) == len(stations)
        assert len(beta_Hperc) == len(stations)

        stats_list: dict[str, dict[str, float]] = {
            stations[i]: dict(
                exceedances_obs=int(exceedances[0][i]),
                MPI_mean=obsmean[i],
                MPI_R_t=MPI_R_t[i],
                MPI_bias_t=MPI_bias_t[i],
                MPI_std_t=MPI_std_t[i],
                MPI_R_s=MPI_R_s,
                MPI_std_s=MPI_std_s,
                MPI_Hperc=beta_Hperc[i],
                bias=bias[i],
                NMB=NMB[i],
                RMSU=βRMSUt[i],
                sign=[sign[i]],
                crms=crms[i],
                rms=[rms[i]],
                beta_mqi=[mqi[i]],
                persistence_model=False,
                station_type=station_types[i],
                **{k: (str(v) if k == "freq" else v) for (k, v) in SPECIES[var_name].items()},
            )
            for i in range(len(stations))
        }

        return stats_list

    @staticmethod
    def _exceedances(data: xr.DataArray, var_name: str) -> list[np.array]:
        if var_name == "concno2":
            new_data = data.resample(time="1D", skipna=True).max()
        else:
            new_data = data

        obsvals = new_data.data[0]
        modvals = new_data.data[1]

        mask = np.isfinite(obsvals) & np.isfinite(modvals)
        obsex = np.sum(obsvals > EXC_THRESHOLDS[var_name], axis=0, where=mask)
        modex = np.sum(modvals > EXC_THRESHOLDS[var_name], axis=0, where=mask)

        return [obsex, modex]

    @staticmethod
    def _NMB(x: np.ndarray, y: np.ndarray) -> np.ndarray:
        num = np.nansum(x - y, axis=0)
        denum = np.nansum(x, axis=0)
        return np.where(denum == 0, np.nan, num / denum)

    @staticmethod
    def pearson_R(x: np.ndarray, y: np.ndarray) -> np.ndarray:
        mask = np.isfinite(x) & np.isfinite(y)

        xmean = np.mean(x, axis=0, where=mask)
        ymean = np.mean(y, axis=0, where=mask)
        xm = x - xmean
        ym = y - ymean
        normxm = np.sqrt(np.sum(xm * xm, axis=0, where=mask))
        normym = np.sqrt(np.sum(ym * ym, axis=0, where=mask))

        r = np.where(
            normxm * normym == 0.0,
            np.nan,
            np.sum(xm * ym, axis=0, where=mask) / (normxm * normym),
        )

        return r

    # in the following _s and _t are for the spatial(ly averaged) and temporal(ly-only averaged) quantities,
    # we define a _βRMSU_t to be used for the calculation of (following MeteoFrance/evaltools naming)
    # Time Bias Norm (_MPI_bias_t), Time Corr Norm (_MPI_r_t) and Time StdDev Norm (_MPI_std_t)
    # and a _βRMSU_s to be used for the calculation of Space Corr Norm (_MPI_R_s) and Space StDev Norm (_MPI_std_s)

    @staticmethod
    def _βRMSU_t(
        obsvals: np.ndarray,
        beta: float,
        var_name: str,
        mask: np.ndarray,
    ) -> np.ndarray:
        def obsuncertainty(obs: np.ndarray, spec: str) -> np.ndarray:
            """eq. 7 and 37 here https://fairmode.jrc.ec.europa.eu/document/fairmode/WG1/Guidance_MQO_Bench_vs3.3_20220519.pdf"""

            if spec not in SPECIES:
                raise ValueError(f"Unsupported {spec=}")

            UrRV = SPECIES[spec]["UrRV"]
            RV = SPECIES[spec]["RV"]
            alpha = SPECIES[spec]["alpha"]

            in_sqrt = (1 - alpha**2) * (obs**2) + alpha**2 * RV**2

            return UrRV * np.sqrt(in_sqrt)

        return beta * np.sqrt(
            np.nanmean(np.square(obsuncertainty(obsvals, var_name), where=mask), axis=0)
        )

    @staticmethod
    def _βRMSU_s(obsmean: np.ndarray, beta: float, var_name: str) -> float:
        def obsuncertainty(obs: np.ndarray, spec: str) -> np.ndarray:
            """eq. 39 here https://fairmode.jrc.ec.europa.eu/document/fairmode/WG1/Guidance_MQO_Bench_vs3.3_20220519.pdf"""

            if spec not in SPECIES:
                raise ValueError(f"Unsupported {spec=}")

            UrRV = SPECIES[spec]["UrRV"]
            RV = SPECIES[spec]["RV"]
            alpha = SPECIES[spec]["alpha"]
            N_p = SPECIES[spec]["Np"]
            N_np = SPECIES[spec]["Nnp"]

            in_sqrt = (1 - alpha**2) / N_p * (obs**2) + alpha**2 * RV**2 / N_np

            return UrRV * np.sqrt(in_sqrt)

        return beta * np.sqrt(np.nanmean(np.square(obsuncertainty(obsmean, var_name))))

    @staticmethod
    def _MPI_R_t(obsstd: np.array, modstd: np.array, R: float, βRMSUt: np.array) -> np.array:
        # TIME Corr Norm: 2 sigma_O sigma_M(1-R) / (beta^2 RMS_U^2)
        # ----------------------------------------------------------
        # MF eq.: (2.*scores['obs_std']*scores['sim_std']*(1. - scores['PearsonR'])) / (beta*rmsu)**2
        return np.where(βRMSUt == 0, np.nan, 2 * modstd * obsstd * (1 - R) / βRMSUt**2)

    @staticmethod
    def _MPI_bias_t(obsmean: np.array, modmean: np.array, βRMSUt: np.array) -> np.array:
        # TIME Bias Norm: |BIAS| / (beta RMS_U)
        # -------------------------------------------
        # MF eq.: scores['MeanBias']/(beta*rmsu)
        return np.where(
            βRMSUt == 0, np.nan, (modmean - obsmean) / βRMSUt
        )  # check the abs here, why NMF does not have it?

    @staticmethod
    def _MPI_std_t(obsstd: np.array, modstd: np.array, βRMSUt: np.array) -> np.array:
        # TIME StDev Norm: (sigma_M-sigma_O) / (beta RMS_U)
        # ---------------------------------------------------
        # MF eq.: (scores['sim_std']-scores['obs_std'])/(beta*rmsu)
        return np.where(βRMSUt == 0, np.nan, (modstd - obsstd) / βRMSUt)

    @staticmethod
    def _MPI_R_s(obsmean: np.array, modmean: np.array, βRMSUs: float) -> float:
        # SPACE Corr Norm: 2 sigma_bar{O} sigma_bar{M} (1-R) / (beta^2 RMS_bar{U}^2)
        # ----------------------------------------------------------------------------
        # MF eq.: ((2.*np.nanstd(obs)*np.nanstd(sim)*(1. - corr)) / (beta*rmsu_)**2)
        # where sim = scores['sim_mean'], obs = scores['obs_mean']
        # corr = ((np.nanmean((obs-np.nanmean(obs))*(sim-np.nanmean(sim)))) / (np.nanstd(obs)*np.nanstd(sim)))
        corr = (
            0.0
            if np.nanstd(obsmean) * np.nanstd(modmean) == 0
            else np.nanmean((obsmean - np.nanmean(obsmean)) * (modmean - np.nanmean(modmean)))
            / (np.nanstd(obsmean) * np.nanstd(modmean))
        )
        return (
            2 * np.nanstd(obsmean) * np.nanstd(modmean) * (1.0 - corr) / βRMSUs**2
            if βRMSUs != 0
            else np.nan
        )

    @staticmethod
    def _MPI_std_s(obsmean: np.array, modmean: np.array, βRMSUs: float) -> float:
        # SPACE StDev Norm: (sigma_bar{M}-sigma_bar{O}) / ( beta RMS_bar{U})
        # -------------------------------------------------------------------
        # MF eq.: (np.nanstd(sim)-np.nanstd(obs))/(beta*rmsu_)
        # where sim = scores['sim_mean'], obs = scores['obs_mean']
        return (np.nanstd(modmean) - np.nanstd(obsmean)) / βRMSUs if βRMSUs != 0 else np.nan

    @staticmethod
    def _fairmode_sign(mod_std: float, obs_std: float, R: float) -> float:
        a = np.where(
            np.logical_or(obs_std <= 0, R >= 1),
            1,
            np.abs(mod_std - obs_std) / (obs_std * np.sqrt(2 * (1 - R))),
        )
        return np.where(a >= 1, 1.0, -1.0)

    @staticmethod
    def _crms(mod_std: float, obs_std: float, R: float) -> float:
        """Returns the Centered Root Mean Squared Error"""
        return np.sqrt(mod_std**2 + obs_std**2 - 2 * mod_std * obs_std * R)

    @staticmethod
    def _mqi(rms: float, rmsu: float, *, beta: float) -> float:
        """Model Quality Indicator (MQI). Pass beta=1 for `beta MQI`"""
        return rms / (rmsu * beta)

    @staticmethod
    def _mb(bias: float, rmsu: float, *, beta: float) -> float:
        """Model Bias(MB). Pass beta=1 for `beta MB`"""
        return bias / (rmsu * beta)

    @staticmethod
    def _beta_Hperc(obs: np.ndarray, mod: np.ndarray, var_name: str, beta=1) -> np.ndarray:
        percentile = SPECIES[var_name]["percentile"]
        Operc = np.nanpercentile(obs, percentile, axis=0)
        Mperc = np.nanpercentile(mod, percentile, axis=0)

        factor = SPECIES[var_name]["alpha"] ** 2 * SPECIES[var_name]["RV"] ** 2
        uncertainty_Operc = SPECIES[var_name]["UrRV"] * np.sqrt(
            (1 - SPECIES[var_name]["alpha"] ** 2) * Operc**2 + factor
        )

        return (Mperc - Operc) / (beta * uncertainty_Operc)



from __future__ import annotations

from configparser import ConfigParser
from pathlib import Path

from pyaerocom.data import resources
from pyaerocom.exceptions import VariableDefinitionError


def parse_variables_ini(fpath: str | Path | None = None):
    """Returns instance of ConfigParser to access information"""

    if fpath is None:
        with resources.path("pyaerocom.data", "variables.ini") as path:
            fpath = path

    if isinstance(fpath, str):
        fpath = Path(fpath)
    if not fpath.exists():
        raise FileNotFoundError(f"FATAL: variables.ini file could not be found at {fpath}")

    parser = ConfigParser()
    parser.read(fpath)
    return parser


def parse_aliases_ini():
    """Returns instance of ConfigParser to access information"""
    with resources.path("pyaerocom.data", "aliases.ini") as path:
        fpath = path

    parser = ConfigParser()
    parser.read(fpath)
    return parser


def _read_alias_ini(parser: ConfigParser | None = None):
    """Read all alias definitions from aliases.ini file and return as dict

    Returns
    -------
    dict
        keys are AEROCOM standard names of variable, values are corresponding
        aliases
    """
    if parser is None:
        parser = parse_aliases_ini()
    aliases = {}
    items = parser["aliases"]
    for var_name in items:
        _aliases = [x.strip() for x in items[var_name].strip().split(",")]
        for alias in _aliases:
            aliases[alias] = var_name
    for var_fam, alias_fam in parser["alias_families"].items():
        if "," in alias_fam:
            raise Exception(
                f"Found invalid definition of alias family {var_fam}: {alias_fam}. "
                f"Only one family can be mapped to a variable name"
            )
    return aliases


def get_aliases(var_name: str, parser: ConfigParser | None = None):
    """Get aliases for a certain variable"""
    if parser is None:
        parser = ConfigParser()
        with resources.path("pyaerocom.data", "aliases.ini") as path:
            parser.read(path)

    info = parser["aliases"]
    aliases = []
    if var_name in info:
        aliases.extend([a.strip() for a in info[var_name].split(",")])
    for var_fam, alias_fam in parser["alias_families"].items():
        if var_name.startswith(var_fam):
            alias = var_name.replace(var_fam, alias_fam)
            aliases.append(alias)
    return aliases


def _check_alias_family(var_name: str, parser: ConfigParser):
    for var_fam, alias_fam in parser["alias_families"].items():
        if var_name.startswith(alias_fam):
            var_name_aerocom = var_name.replace(alias_fam, var_fam)
            return var_name_aerocom
    raise VariableDefinitionError(
        "Input variable could not be identified as "
        "belonging to either of the available alias "
        "variable families"
    )


def get_variable(var_name: str):
    """
    Get a certain variable

    Parameters
    ----------
    var_name : str
        variable name

    Returns
    -------
    Variable
    """
    from pyaerocom import const

    return const.VARS[var_name]



from __future__ import annotations

import sys

from pyaerocom.data import resources

if sys.version_info >= (3, 11):  # pragma: no cover
    import tomllib
else:  # pragma: no cover
    import tomli as tomllib

_VARIABLES = "uemep_variables.toml"


def uemep_variables() -> dict[str, str]:
    """Read variable definitions from uemep_variables.yaml file

    Returns
    -------
    dict
        keys are AEROCOM standard names of variable, values are uEMEP variables
    """
    assert resources.is_resource(__package__, _VARIABLES), f"{_VARIABLES} missing in {__package__}"
    variables = tomllib.loads(resources.read_text(__package__, _VARIABLES))
    return variables["uemep_variables"]



from __future__ import annotations

from pathlib import Path

import numpy as np
import pytest
import xarray as xr
from numpy.typing import ArrayLike
from pydantic import ValidationError

from pyaerocom import ColocatedData
from pyaerocom.config import ALL_REGION_NAME
from pyaerocom.exceptions import DataCoverageError, DataDimensionError
from tests.fixtures.collocated_data import EXAMPLE_FILE


@pytest.mark.parametrize("data", [EXAMPLE_FILE, str(EXAMPLE_FILE), np.ones((2, 3, 4))])
def test_ColocatedData_initialization(data: Path | str | ArrayLike):
    cd = ColocatedData(data=data)
    assert isinstance(cd.data, xr.DataArray)


@pytest.mark.parametrize(
    "data,exception",
    [
        ("Blaaaaa", ValueError),
        (np.ones(3), ValidationError),
        ({}, ValueError),
        (xr.DataArray(np.ones((2, 2, 2, 2, 1))), ValidationError),
    ],
)
def test_ColocatedData_initialization_error(data, exception: type[Exception]):
    with pytest.raises(exception):
        ColocatedData(data=data).data


def test_ColocatedData__init__():
    # test that ColocatedData can be given positional arguments
    data = np.ones((2, 2, 1))
    col_data = ColocatedData(data)
    assert col_data


def test_ColocatedData_data():
    data = xr.DataArray(np.ones((2, 2, 1)))
    col = ColocatedData(data=data)
    assert col.data is data


def test_ColocatedData_data_error():
    col = ColocatedData()
    with pytest.raises(ValueError):
        col.data = "Blaaa"


@pytest.mark.parametrize("coldataset", ["tm5_aeronet"])
def test_ColocatedData_data_source(coldata: ColocatedData):
    ds = coldata.data_source
    assert isinstance(ds, xr.DataArray)
    assert len(ds) == 2


@pytest.mark.parametrize("coldataset", ["fake_nodims"])
def test_ColocatedData_data_source_error(coldata: ColocatedData):
    with pytest.raises(AttributeError) as e:
        coldata.data_source
    assert str(e.value).endswith("has no attribute 'data_source'")


@pytest.mark.parametrize("coldataset", ["tm5_aeronet"])
def test_ColocatedData_var_name(coldata: ColocatedData):
    assert isinstance(coldata.var_name, list)


@pytest.mark.parametrize("coldataset", ["fake_nodims"])
def test_ColocatedData_var_name_error(coldata: ColocatedData):
    with pytest.raises(AttributeError) as e:
        coldata.var_name
    assert str(e.value).endswith("has no attribute 'var_name'")


@pytest.mark.parametrize("coldataset", ["tm5_aeronet"])
def test_ColocatedData_latitude(coldata: ColocatedData):
    assert isinstance(coldata.latitude, xr.DataArray)


@pytest.mark.parametrize("coldataset", ["fake_nodims"])
def test_ColocatedData_latitude_error(coldata: ColocatedData):
    with pytest.raises(AttributeError) as e:
        coldata.latitude
    assert str(e.value).endswith("object has no attribute 'latitude'")


@pytest.mark.parametrize("coldataset", ["tm5_aeronet"])
def test_ColocatedData_longitude(coldata: ColocatedData):
    assert isinstance(coldata.longitude, xr.DataArray)


@pytest.mark.parametrize("coldataset", ["fake_nodims"])
def test_ColocatedData_longitude_error(coldata: ColocatedData):
    with pytest.raises(AttributeError) as e:
        coldata.longitude
    assert str(e.value).endswith("object has no attribute 'longitude'")


@pytest.mark.parametrize("coldataset", ["tm5_aeronet"])
def test_ColocatedData_time(coldata: ColocatedData):
    assert isinstance(coldata.time, xr.DataArray)


@pytest.mark.parametrize("coldataset", ["fake_nodims"])
def test_ColocatedData_time_error(coldata: ColocatedData):
    with pytest.raises(AttributeError) as e:
        coldata.time
    assert str(e.value).endswith("object has no attribute 'time'")


@pytest.mark.parametrize(
    "coldataset,lat_range",
    [
        ("tm5_aeronet", (-43.2, 43.9)),
        ("fake_4d", (30, 50)),
    ],
)
def test_ColocatedData_lat_range(coldata: ColocatedData, lat_range: tuple[float, float]):
    assert coldata.lat_range == pytest.approx(lat_range, rel=1e-1)


@pytest.mark.parametrize("coldataset", ["fake_nodims"])
def test_ColocatedData_lat_range_error(coldata: ColocatedData):
    with pytest.raises(AttributeError) as e:
        coldata.lat_range
    assert "object has no attribute" in str(e.value)


@pytest.mark.parametrize(
    "coldataset,lon_range",
    [
        ("tm5_aeronet", (-65.3, 121.5)),
        ("fake_4d", (10, 20)),
    ],
)
def test_ColocatedData_lon_range(coldata: ColocatedData, lon_range: tuple[float, float]):
    assert coldata.lon_range == pytest.approx(lon_range, rel=1e-1)


@pytest.mark.parametrize("coldataset", ["fake_nodims"])
def test_ColocatedData_lon_range_error(coldata: ColocatedData):
    with pytest.raises(AttributeError) as e:
        coldata.lon_range
    assert "object has no attribute" in str(e.value)


@pytest.mark.parametrize("coldataset", ["tm5_aeronet"])
def test_ColocatedData_ts_type(coldata: ColocatedData):
    assert isinstance(coldata.ts_type, str)


@pytest.mark.parametrize("coldataset", ["fake_nodims"])
def test_ColocatedData_ts_type_error(coldata: ColocatedData):
    with pytest.raises(ValueError) as e:
        coldata.ts_type
    assert str(e.value).endswith("does not contain information about temporal resolution")


@pytest.mark.parametrize("coldataset", ["tm5_aeronet"])
def test_ColocatedData_units(coldata: ColocatedData):
    units = coldata.units
    assert isinstance(units, list)
    assert bool(units) and all(isinstance(x, str) for x in units)


@pytest.mark.parametrize("coldataset", ["fake_nodims"])
def test_ColocatedData_units_error(coldata: ColocatedData):
    with pytest.raises(KeyError) as e:
        coldata.units
    assert "units" in str(e.value)


@pytest.mark.parametrize(
    "coldataset,num_coords",
    [
        ("fake_4d", 6),
        ("tm5_aeronet", 8),
        ("fake_3d", 4),
    ],
)
def test_ColocatedData_num_coords(coldata: ColocatedData, num_coords: int):
    assert coldata.num_coords == num_coords


@pytest.mark.parametrize("coldataset", ["fake_nodims"])
def test_ColocatedData_num_coords_error(coldata: ColocatedData):
    with pytest.raises(DataDimensionError) as e:
        coldata.num_coords
    assert str(e.value).startswith("Need dimension")


@pytest.mark.parametrize(
    "coldataset,num_coords_with_data",
    [
        ("tm5_aeronet", 8),
        ("fake_3d", 4),
        ("fake_4d", 5),
    ],
)
def test_ColocatedData_num_coords_with_data(coldata: ColocatedData, num_coords_with_data: int):
    assert coldata.num_coords_with_data == num_coords_with_data


@pytest.mark.parametrize(
    "coldataset,error",
    [
        ("fake_nodims", "Need dimension"),
    ],
)
def test_ColocatedData_num_coords_with_data_error(coldata: ColocatedData, error: str):
    with pytest.raises(DataDimensionError) as e:
        coldata.num_coords_with_data
    assert error in str(e.value)


@pytest.mark.parametrize(
    "coldataset,num_coords",
    [
        ("tm5_aeronet", 8),
        ("fake_4d", 5),
    ],
)
def test_ColocatedData_get_coords_valid_obs(coldata: ColocatedData, num_coords: int):
    coords_valid_obs = coldata.get_coords_valid_obs()
    assert isinstance(coords_valid_obs, list)
    assert len(coords_valid_obs) == 2
    assert all(len(coord) == num_coords for coord in coords_valid_obs)


@pytest.mark.parametrize("coldataset", ["fake_nodims"])
def test_ColocatedData_get_coords_valid_obs_error(coldata: ColocatedData):
    with pytest.raises(ValueError) as e:
        coldata.get_coords_valid_obs()
    assert str(e.value).startswith("'time' not found in array dimensions")


@pytest.mark.parametrize(
    "coldataset,use_area_weights,chk",
    [
        ("tm5_aeronet", False, {"nmb": -0.129, "R": 0.853}),
        # has random numbers in it so nmb, R check is risky with rtol=1e-2
        ("fake_3d", False, {"num_coords_with_data": 4}),
        ("fake_4d", False, {"nmb": 0}),
        ("fake_4d", True, {"nmb": 0}),
    ],
)
def test_ColocatedData_calc_statistics(coldata: ColocatedData, use_area_weights: bool, chk: dict):
    output = coldata.calc_statistics(use_area_weights=use_area_weights)
    assert isinstance(output, dict)
    for key, val in chk.items():
        assert output[key] == pytest.approx(val, rel=1e-2, nan_ok=True)


@pytest.mark.parametrize("coldataset", ["fake_nodims"])
def test_ColocatedData_calc_statistics_error(coldata: ColocatedData):
    with pytest.raises(DataDimensionError) as e:
        coldata.calc_statistics()
    assert str(e.value).startswith("Need dimension")


@pytest.mark.parametrize(
    "coldataset,aggr,statistics",
    [
        ("tm5_aeronet", None, {"nmb": -0.065, "R": 0.679}),
        ("fake_3d", None, {}),
        ("fake_4d", None, {"nmb": 0}),
        ("tm5_aeronet", "median", {"nmb": -0.0136, "R": 0.851}),
    ],
)
def test_ColocatedData_calc_temporal_statistics(
    coldata: ColocatedData, aggr: str | None, statistics: dict[str, float]
):
    temporal_statistics = coldata.calc_temporal_statistics(aggr=aggr)
    assert isinstance(temporal_statistics, dict)
    for key, val in statistics.items():
        assert temporal_statistics[key] == pytest.approx(val, rel=1e-2)


@pytest.mark.parametrize(
    "coldataset,aggr,exception,error",
    [
        ("fake_nodims", None, DataDimensionError, "Need dimension"),
        ("tm5_aeronet", "max", ValueError, "So far only mean and median are supported"),
    ],
)
def test_ColocatedData_calc_temporal_statistics_error(
    coldata: ColocatedData, aggr: str | None, exception: type[Exception], error: str
):
    with pytest.raises(exception) as e:
        coldata.calc_temporal_statistics(aggr=aggr)
    assert str(e.value).startswith(error)


@pytest.mark.parametrize(
    "coldataset,args,statistics",
    [
        ("tm5_aeronet", {}, {"nmb": -0.304, "R": 0.893}),
        ("fake_3d", {}, {}),
        ("fake_4d", {}, {"nmb": 0}),
        ("fake_4d", {"use_area_weights": True}, {"nmb": 0}),
        ("tm5_aeronet", {"aggr": "median"}, {"nmb": -0.42, "R": 0.81}),
    ],
)
def test_ColocatedData_calc_spatial_statistics(
    coldata: ColocatedData, args: dict, statistics: dict[str, float]
):
    spatial_statistics = coldata.calc_spatial_statistics(**args)
    assert isinstance(spatial_statistics, dict)
    for key, val in statistics.items():
        assert spatial_statistics[key] == pytest.approx(val, rel=1e-2)


@pytest.mark.parametrize(
    "coldataset,aggr,exception,error",
    [
        ("fake_nodims", None, DataDimensionError, "Need dimension"),
        ("tm5_aeronet", "max", ValueError, "So far only mean and median are supported"),
    ],
)
def test_ColocatedData_calc_spatial_statistics_error(
    coldata: ColocatedData, aggr: str | None, exception: type[Exception], error: str
):
    with pytest.raises(exception) as e:
        coldata.calc_spatial_statistics(aggr=aggr)
    assert str(e.value).startswith(error)


def test_meta_access_filename():
    name = f"od550bc_ang4487aer_MOD-AEROCOM-MEDIAN_REF-42AeronET_20000101_20201231_monthly_{ALL_REGION_NAME}-noMOUNTAINS.nc"

    meta = {
        "model_var": "od550bc",
        "obs_var": "ang4487aer",
        "model_name": "AEROCOM-MEDIAN",
        "obs_name": "42AeronET",
        "start": "20000101",
        "stop": "20201231",
        "ts_type": "monthly",
        "filter_name": f"{ALL_REGION_NAME}-noMOUNTAINS",
    }

    _meta = ColocatedData.get_meta_from_filename(name)
    assert _meta == meta


def test_read_colocated_data(coldata_tm5_aeronet):
    loaded = ColocatedData(data=EXAMPLE_FILE)
    mean_loaded = np.nanmean(loaded.data)
    mean_fixture = np.nanmean(coldata_tm5_aeronet.data.data)
    assert mean_fixture == mean_loaded


@pytest.mark.parametrize(
    "region_id,latrange,lonrange,numst",
    [
        ("RBU", (29.45, 66.26), (22, -170), 2),  # crosses lon=180 border
        (ALL_REGION_NAME, (-90, 90), (-180, 180), 8),
        ("NHEMISPHERE", (0, 90), (-180, 180), 5),
        ("EUROPE", (40, 72), (-10, 40), 2),
        ("OCN", (-90, 90), (-180, 180), 8),
    ],
)
def test_apply_latlon_filter(
    coldata_tm5_aeronet,
    region_id: str,
    latrange: tuple[float, float],
    lonrange: tuple[float, float],
    numst: int,
):
    filtered = coldata_tm5_aeronet.apply_latlon_filter(region_id=region_id)

    lats, lons = filtered.data.latitude.data, filtered.data.longitude.data
    assert len(filtered.data.station_name.data) == numst
    if numst > 0:
        assert lats.min() > latrange[0]
        assert lats.max() < latrange[1]
        if lonrange[0] < lonrange[1]:
            assert lons.min() > lonrange[0]
            assert lons.max() < lonrange[1]
        else:
            assert -180 < lons.min() < lonrange[1] or lonrange[0] < lons.min() < 180
            assert -180 < lons.max() < lonrange[1] or lonrange[0] < lons.max() < 180


@pytest.mark.parametrize(
    "region_id",
    [
        "PAN",  # crosses lon=180 border
        "NAM",  # crosses lon=180 border
    ],
)
def test_apply_latlon_filter_error(coldata_tm5_aeronet, region_id: str):
    with pytest.raises(DataCoverageError):
        coldata_tm5_aeronet.apply_latlon_filter(region_id=region_id)


@pytest.mark.parametrize(
    "coldataset,region_id,check_country_meta,latrange,lonrange,numst",
    [
        ("fake_4d", "EUROPE", False, (40, 72), (-10, 40), 4),
        ("tm5_aeronet", "NHEMISPHERE", False, (0, 90), (-180, 180), 5),
        ("tm5_aeronet", "EUROPE", False, (40, 72), (-10, 40), 2),
        ("tm5_aeronet", "OCN", False, (-59.95, 66.25), (-132.55, 119.95), 1),
        ("tm5_aeronet", "Brazil", True, (-59.95, 66.25), (-132.55, 119.95), 1),
    ],
)
def test_ColocatedData_filter_region(
    coldata: ColocatedData,
    region_id: str,
    check_country_meta: bool,
    latrange: tuple[float, float],
    lonrange: tuple[float, float],
    numst: int,
):
    coldata = coldata
    if check_country_meta:
        coldata.check_set_countries()

    filtered = coldata.filter_region(region_id=region_id, check_country_meta=check_country_meta)
    lats, lons = filtered.data.latitude.data, filtered.data.longitude.data
    assert lats.min() >= latrange[0]
    assert lats.max() <= latrange[1]
    assert lons.min() >= lonrange[0]
    assert lons.max() <= lonrange[1]
    assert filtered.num_coords == numst


@pytest.mark.parametrize("coldataset", ["fake_4d"])
def test_ColocatedData_filter_region_error(coldata: ColocatedData):
    with pytest.raises(DataDimensionError) as e:
        coldata.check_set_countries()
    assert str(e.value).startswith("Countries cannot be assigned")

    with pytest.raises(AttributeError) as e:
        coldata.filter_region(region_id="France", check_country_meta=True)
    assert str(e.value).endswith("'ColocatedData' object has no attribute 'countries_available'")


@pytest.mark.parametrize(
    "coldataset,filename",
    [
        (
            "tm5_aeronet",
            f"od550aer_od550aer_MOD-TM5_AP3-CTRL2016_REF-AeronetSunV3L2Subset.daily_20100115_20101215_monthly_{ALL_REGION_NAME}-noMOUNTAINS.nc",
        ),
        (
            "fake_3d_hr",
            f"vmro3_vmro3_MOD-fakemod_REF-fakeobs_20180110_20180117_hourly_{ALL_REGION_NAME}-wMOUNTAINS.nc",
        ),
        (
            "fake_3d",
            f"concpm10_concpm10_MOD-fakemod_REF-fakeobs_20000115_20191215_monthly_{ALL_REGION_NAME}-wMOUNTAINS.nc",
        ),
    ],
)
def test_ColocatedData_to_netcdf(coldata: ColocatedData, tmp_path: Path, filename: str):
    path = tmp_path / filename
    assert not path.exists()
    file = coldata.to_netcdf(tmp_path)
    assert path == Path(file)
    assert path.exists()


@pytest.mark.parametrize("coldataset", ["tm5_aeronet", "fake_3d_hr", "fake_3d"])
def test_ColocatedData_read_netcdf(coldata: ColocatedData, tmp_path: Path):
    file = coldata.to_netcdf(tmp_path)
    assert Path(file).exists()
    cd = ColocatedData(file)
    assert isinstance(cd, ColocatedData)


@pytest.mark.parametrize(
    "coldataset,args,mean",
    [
        ("tm5_aeronet", dict(to_ts_type="yearly"), 0.336),
        ("tm5_aeronet", dict(to_ts_type="yearly", min_num_obs=14), np.nan),
        ("tm5_aeronet", dict(to_ts_type="yearly", settings_from_meta=True), 0.336),
        ("tm5_aeronet", dict(to_ts_type="yearly", colocate_time=True), 0.363),
    ],
)
def test_ColocatedData_resample_time(coldata: ColocatedData, args: dict, mean):
    resampled = coldata.resample_time(**args)
    assert (resampled is coldata) == args.get("inplace", False)

    resampled_mean = resampled.data.mean().data
    assert resampled_mean == pytest.approx(mean, abs=1e-3, nan_ok=True)


@pytest.mark.parametrize(
    "coldataset",
    (
        pytest.param(
            "tm5_aeronet",
        ),
        pytest.param(
            "fake_3d_hr",
        ),
        pytest.param(
            "fake_3d",
        ),
    ),
)
def test_ColocatedData_to_dataframe(coldata: ColocatedData):
    df = coldata.to_dataframe()

    exp_columns = set(
        [
            "time",
            "station_name",
            "data_source_obs",
            "latitude",
            "longitude",
            "altitude",
            f"{coldata.var_name[0]}_obs",
            "data_source_mod",
            f"{coldata.var_name[1]}_mod",
        ]
    )

    assert df.shape[1] == 9
    assert set(df.columns) == exp_columns
    assert not df["time"].isnull().values.any()
    assert not df["data_source_obs"].isnull().values.any()
    assert not df["data_source_mod"].isnull().values.any()


@pytest.mark.parametrize(
    "coldataset",
    (
        pytest.param(
            "fake_4d",
        ),
    ),
)
def test_ColocatedData_to_dataframe_exception(coldata: ColocatedData):
    with pytest.raises(NotImplementedError):
        coldata.to_dataframe()


@pytest.mark.parametrize(
    "coldataset",
    (
        pytest.param(
            "tm5_aeronet",
        ),
        pytest.param(
            "fake_3d_hr",
        ),
        pytest.param(
            "fake_3d",
        ),
    ),
)
def test_ColocatedData_from_dataframe(coldata: ColocatedData):
    df = coldata.to_dataframe()

    ColocatedData.from_dataframe(df)



from pyaerocom.units.helpers import get_standard_unit
import logging

logger = logging.getLogger(__name__)


def harmonise_units(data, data_ref, *, var: str, var_ref: str, inplace=False) -> tuple:
    """Tries to perform unit conversion for the provided data object, so that
    the units of the data objects match for the given variable names.

    If var == var_ref, both data objects will be converted to the unit provided
    by pyaerocom.units.get_standard_unit.

    If var != var_ref, both data objects will be attempted converted to the unit
    provided by pyaerocom.units.get_standard_unit for var_ref.

    :param data: Data
    :param data: Reference data.
    :param var: Varname for data.
    :param var_ref: Varname for data_ref.
    :param inplace: Whether to perform conversion inplace.
    :return: tuple of length n, containing converted data and ref_data.
    """
    std_unit = get_standard_unit(var)
    std_unit_ref = get_standard_unit(var_ref)

    if std_unit != std_unit_ref:
        logger.info(
            f"'std_unit' ({std_unit}) and 'std_unit_ref' ({std_unit_ref}) do not match. Attempting to convert both data objects to {std_unit_ref}."
        )

    harmonised_unit = std_unit_ref

    try:
        data = data.convert_unit(harmonised_unit, inplace=inplace)
    except AttributeError:
        data = data.check_convert_var_units(var, to_unit=harmonised_unit, inplace=inplace)

    try:
        data_ref = data_ref.convert_unit(harmonised_unit, inplace=inplace)
    except AttributeError:
        data_ref = data_ref.check_convert_var_units(
            var_ref, to_unit=harmonised_unit, inplace=inplace
        )

    return data, data_ref



import numpy as np
from scipy.stats import ConstantInputWarning, kendalltau, spearmanr

from pyaerocom.mathutils import corr, sum

from .._warnings import ignore_warnings


def stat_R(data: np.ndarray, ref_data: np.ndarray, weights: np.ndarray | None) -> np.float64:
    """
    Pearson correlation coefficient implementation.
    """
    return corr(data, ref_data, weights)


def stat_rms(data: np.ndarray, ref_data: np.ndarray, weights: np.ndarray | None) -> np.float64:
    """
    Root mean square implementation.
    """
    difference = data - ref_data
    return np.sqrt(np.average(difference**2, weights=weights))


def stat_mb(data: np.ndarray, ref_data: np.ndarray, weights: np.ndarray | None) -> np.float64:
    """
    Mean bias implementation.
    """
    difference = data - ref_data
    return sum(difference, weights=weights) / len(data)


def stat_nmb(data: np.ndarray, ref_data: np.ndarray, weights: np.ndarray | None) -> np.float64:
    """
    Normalised mean bias implementation.
    """
    sum_ref_data = sum(ref_data, weights=weights)
    if sum_ref_data == 0:
        return 0
    return sum(data - ref_data, weights) / sum_ref_data


def stat_mab(data: np.ndarray, ref_data: np.ndarray, weights: np.ndarray | None) -> np.float64:
    """
    Mean absolute bias implementation.
    """
    difference = data - ref_data
    return sum(np.abs(difference)) / len(data)


@ignore_warnings(RuntimeWarning, "invalid value encountered in divide")
def stat_mnmb(data: np.ndarray, ref_data: np.ndarray, weights: np.ndarray | None) -> np.float64:
    """
    Modified normalised mean bias implementation.
    """
    difference = data - ref_data
    if np.all(np.isnan(difference)):
        return np.nan

    return 2 / len(data) * sum(difference / (data + ref_data))


@ignore_warnings(RuntimeWarning, "invalid value encountered in divide")
def stat_fge(data: np.ndarray, ref_data: np.ndarray, weights: np.ndarray | None) -> np.float64:
    """
    Fractional gross error implementation.
    """
    difference = data - ref_data
    if np.all(np.isnan(difference)):
        return np.nan

    return 2 / len(data) * sum(np.abs(difference / (data + ref_data)), weights=weights)


@ignore_warnings(
    ConstantInputWarning, "An input array is constant; the correlation coefficient is not defined."
)
def stat_R_spearman(
    data: np.ndarray, ref_data: np.ndarray, weights: np.ndarray | None
) -> np.float64:
    """
    Spearman corr. coefficient implementation.
    """
    return spearmanr(data, ref_data)[0]


def stat_R_kendall(
    data: np.ndarray, ref_data: np.ndarray, weights: np.ndarray | None
) -> np.float64:
    """
    Kendall's tau implementation.
    """
    return kendalltau(data, ref_data)[0]



import logging
import os
from pathlib import Path

from pydantic import BaseModel

from pyaerocom import const
from pyaerocom.aeroval.modelentry import ModelEntry
from pyaerocom.griddeddata import GriddedData
from pyaerocom.units.datetime import get_highest_resolution, to_pandas_timestamp
from pyaerocom.helpers import (
    get_max_period_range,
    make_dummy_cube,
    start_stop,
)
from pyaerocom.variable import Variable

logger = logging.getLogger(__name__)


def _check_statistics_periods(periods: list) -> list:
    """
    Check input list of period strings is valid

    Parameters
    ----------
    periods : list
        list containing period strings to be checked.

    Raises
    ------
    ValueError
        if input is not a list or any of the provided periods in that list is
        not a string or invalid.

    Returns
    -------
    list
        list of periods

    """
    checked = []
    if not isinstance(periods, list):
        raise ValueError("statistics_periods needs to be a list")
    for per in periods:
        if not isinstance(per, str):
            raise ValueError("All periods need to be strings")
        spl = [x.strip() for x in per.split("-")]
        # periods can be also dates or date ranges since cams2_83
        if len(spl) == 2:
            if len(spl[0]) != len(spl[1]):
                raise ValueError(f"{spl[0]} not on the same format as {spl[1]}")

        if len(spl) > 2:
            raise ValueError(
                f"Invalid value for period ({per}), can be either single "
                f"years/dates or range of years/dates (e.g. 2000-2010)."
            )
        years = True if len(spl[0]) == 4 else False
        if years:
            _per = "-".join([str(int(val)) for val in spl])
        else:
            # slash in the period string here is required by the aeroval web server logic
            _per = "-".join([to_pandas_timestamp(val).strftime("%Y/%m/%d") for val in spl])
        checked.append(_per)

    return checked


def _period_str_to_timeslice(period: str) -> slice:
    """
    Convert input period to a time slice

    Parameters
    ----------
    period : str
        period, e.g. "2000-2010"

    Raises
    ------
    ValueError
        if input period is invalid

    Returns
    -------
    slice
        slice containing start and end strings.
    """
    spl = period.split("-")
    if len(spl) == 1:
        return slice(spl[0], spl[0])
    elif len(spl) == 2:
        return slice(*spl)
    raise ValueError(period)


def _get_min_max_year_periods(statistics_periods):
    """Get lowest and highest available year from all periods

    Parameters
    ----------
    statistics_periods : list
        list of periods for experiment

    Returns
    -------
    pd.Timestamp
        start year
    pd.Timestamp
        stop year (may be the same as start year, e.g. if periods suggest
        single year analysis).
    """
    startyr, stopyr = start_stop("2100", "1900")
    for per in statistics_periods:
        sl = _period_str_to_timeslice(per)
        perstart, perstop = start_stop(sl.start, sl.stop)
        if perstart < startyr:
            startyr = perstart
        if perstop > stopyr:
            stopyr = perstop
    return startyr, stopyr


def check_if_year(periods: list[str]) -> bool:
    """
    Checks if the periods in the periods list are years or dates
    """
    years = []
    for per in periods:
        spl = [x.strip() for x in per.split("-")]
        if len(spl) == 2:
            if len(spl[0]) != len(spl[1]):
                raise ValueError(f"{spl[0]} not on the same format as {spl[1]}")

        if len(spl) > 2:
            raise ValueError(
                f"Invalid value for period ({per}), can be either single "
                f"years/dates or range of years/dates (e.g. 2000-2010)."
            )
        years.append(True if len(spl[0]) == 4 else False)

    if len(set(years)) != 1:
        raise ValueError(f"Found mix of years and dates in {periods}")
    return list(set(years))[0]


def make_dummy_model(obs_list: list, cfg) -> str:
    # Sets up variable for the model register
    tmpdir = const.LOCAL_TMP_DIR
    const.add_data_search_dir(tmpdir)

    model_id = "dummy_model"
    outdir = os.path.join(tmpdir, f"{model_id}/renamed")

    os.makedirs(outdir, exist_ok=True)

    # Finds dates and freq to use, so that all observations are covered
    (start, stop) = get_max_period_range(cfg.time_cfg.periods)
    freq = get_highest_resolution(*cfg.time_cfg.freqs)

    tmp_var_obj = Variable()
    # Loops over variables in obs
    for obs in obs_list:
        for var in cfg.obs_cfg.get_entry(obs).obs_vars:
            # Create dummy cube

            dummy_cube = make_dummy_cube(var, start_yr=start, stop_yr=stop, freq=freq)

            # Converts cube to GriddedData
            dummy_grid = GriddedData(dummy_cube)

            # Set the value to be the mean of acceptable values to prevent incorrect outlier removal
            # This needs some care though because the defaults are (currently) -inf and inf, which leads to erroneous removal

            if not (
                dummy_grid.var_info.minimum == tmp_var_obj.VMIN_DEFAULT
                or dummy_grid.var_info.maximum == tmp_var_obj.VMAX_DEFAULT
            ):
                dummy_grid.data *= (dummy_grid.var_info.minimum + dummy_grid.var_info.maximum) / 2

            # Loop over each year
            yr_gen = dummy_grid.split_years()

            for dummy_grid_yr in yr_gen:
                # Add to netcdf
                yr = dummy_grid_yr.years_avail()[0]
                vert_code = cfg.obs_cfg.get_entry(obs).obs_vert_type

                save_name = dummy_grid_yr.aerocom_savename(model_id, var, vert_code, yr, freq)
                dummy_grid_yr.to_netcdf(outdir, savename=save_name)

    # Add dummy model to cfg
    cfg.model_cfg.add_entry("dummy", ModelEntry(model_id="dummy_model", model_data_dir=outdir))

    return model_id


def delete_dummy_model(model_id: str) -> None:
    tmpdir = const.LOCAL_TMP_DIR
    const.add_data_search_dir(tmpdir)

    renamed = Path(tmpdir) / f"{model_id}/renamed"
    for path in renamed.glob("*.nc"):
        print(f"Deleting dummy model {path}")
        path.unlink()


class BoundingBox(BaseModel):
    west: float
    east: float
    south: float
    north: float



import numpy as np
import pandas as pd
import pytest

from pyaerocom import ColocatedData, Filter
from pyaerocom.config import ALL_REGION_NAME

from .data_access import DataForTests

CHECK_PATHS = f"coldata/od550aer_REF-AeronetSunV3L2Subset.daily_MOD-TM5_AP3-CTRL2016_20100101_20101231_monthly_{ALL_REGION_NAME}-noMOUNTAINS.nc"
EXAMPLE_FILE = DataForTests(CHECK_PATHS).path


def _create_fake_coldata_3d():
    var = "concpm10"
    filter_name = f"{ALL_REGION_NAME}-wMOUNTAINS"
    regfilter = Filter(name=filter_name)

    dtime = pd.date_range("2000-01-01", "2019-12-31", freq="MS") + np.timedelta64(14, "D")

    lats = [-80, 0, 70, 0.1]
    lons = [-150, 0, 100, 0.1]
    alts = [0, 100, 2000, 10]

    timenum = len(dtime)
    statnum = len(lats)
    c = 1
    statnames = []
    for lat in lats:
        statnames.append(f"FakeStation{c}")
        c += 1
    data = np.ones((2, timenum, statnum))
    xrange_modulation = np.linspace(0, np.pi * 40, timenum)
    data[1] += 0.1  # +10% model bias

    # 1. SITE: modify first site (sin , cos waves)
    data[0, :, 0] += np.sin(xrange_modulation)
    data[1, :, 0] += np.cos(xrange_modulation)  # phase shifted to obs

    years = dtime.year.values
    # 2. SITE: modify second site (yearly stepwise increase, double as
    # pronounced in obs than in model)
    c = 0
    for year in np.unique(years):
        mask = years == year
        data[0, mask, 1] += c * 0.4
        data[1, mask, 1] += c * 0.2

    # set the first and last 10 months of obs to NaN to violate 25% coverage
    # constraint of first and last year to test option annual_stats_constrained
    # in json conversion routines
    data[0, :10, 1] = np.nan
    data[0, -10:, 1] = np.nan

    # 3 SITE: add noise to model and obs
    data[0, :, 2] += np.random.rand(timenum)
    data[1, :, 2] += np.random.rand(timenum)

    meta = {
        "data_source": ["fakeobs", "fakemod"],
        "var_name": [var, var],
        "ts_type": "monthly",  # will be updated below if resampling
        "filter_name": filter_name,
        "ts_type_src": ["monthly", "daily"],
        "start_str": dtime[0].strftime("%Y%m%d"),
        "stop_str": dtime[-1].strftime("%Y%m%d"),
        "var_units": ["ug m-3", "ug m-3"],
        "vert_scheme": "surface",
        "data_level": 3,
        "revision_ref": "20210409",
        "from_files": [],
        "from_files_ref": None,
        "stations_ignored": None,
        "colocate_time": False,
        "obs_is_clim": False,
        "pyaerocom": "0.11.0",
        "min_num_obs": dict(monthly=dict(daily=15), daily=dict(hourly=12)),
        "resample_how": dict(monthly=dict(daily="sum"), daily=dict(hourly="max")),
    }

    meta.update(regfilter.to_dict())

    # create coordinates of DataArray
    coords = {
        "data_source": meta["data_source"],
        "time": dtime,
        "station_name": statnames,
        "latitude": ("station_name", lats),
        "longitude": ("station_name", lons),
        "altitude": ("station_name", alts),
    }

    dims = ["data_source", "time", "station_name"]
    cd = ColocatedData(data=data, coords=coords, dims=dims, name=var, attrs=meta)

    return cd


def _create_fake_trends_coldata_3d():
    var = "concpm10"
    filter_name = f"{ALL_REGION_NAME}-wMOUNTAINS"
    regfilter = Filter(name=filter_name)

    dtime = pd.date_range("2000-01-01", "2019-12-31", freq="MS") + np.timedelta64(14, "D")

    lats = [-80, 0, 70, 0.1]
    lons = [-150, 0, 100, 0.1]
    alts = [0, 100, 2000, 10]

    timenum = len(dtime)
    statnum = len(lats)

    statnames = [f"FakeStation{c}" for c in range(statnum)]

    data = np.ones((2, timenum, statnum))

    trend_slops = [1, 2, 50, -3]

    xs = np.linspace(0, 20, timenum)
    for i, s in enumerate(trend_slops):
        data[0, :, i] = s * xs
        data[1, :, i] = s * xs

    meta = {
        "data_source": ["fakeobs", "fakemod"],
        "var_name": [var, var],
        "ts_type": "monthly",  # will be updated below if resampling
        "filter_name": filter_name,
        "ts_type_src": ["monthly", "daily"],
        "start_str": dtime[0].strftime("%Y%m%d"),
        "stop_str": dtime[-1].strftime("%Y%m%d"),
        "var_units": ["ug m-3", "ug m-3"],
        "vert_scheme": "surface",
        "data_level": 3,
        "revision_ref": "20210409",
        "from_files": [],
        "from_files_ref": None,
        "stations_ignored": None,
        "colocate_time": False,
        "obs_is_clim": False,
        "pyaerocom": "0.11.0",
        "min_num_obs": dict(monthly=dict(daily=15), daily=dict(hourly=12)),
        "resample_how": dict(monthly=dict(daily="sum"), daily=dict(hourly="max")),
    }

    meta.update(regfilter.to_dict())

    # create coordinates of DataArray
    coords = {
        "data_source": meta["data_source"],
        "time": dtime,
        "station_name": statnames,
        "latitude": ("station_name", lats),
        "longitude": ("station_name", lons),
        "altitude": ("station_name", alts),
    }

    dims = ["data_source", "time", "station_name"]
    cd = ColocatedData(data=data, coords=coords, dims=dims, name=var, attrs=meta)

    return cd


def _create_fake_coldata_3d_hourly():
    var = "vmro3"
    filter_name = f"{ALL_REGION_NAME}-wMOUNTAINS"
    regfilter = Filter(name=filter_name)

    dtime = pd.date_range("2018-01-10T00:00:00", "2018-01-17T23:59:00", freq="h")

    lats = [-80]
    lons = [-150]
    alts = [10]

    timenum = len(dtime)
    statnum = len(lats)
    c = 1
    statnames = []
    for lat in lats:
        statnames.append(f"FakeStation{c}")
        c += 1
    data = np.ones((2, timenum, statnum))
    xrange_modulation = np.linspace(0, np.pi * 40, timenum)
    signal = np.sin(xrange_modulation)
    data[1] += 0.1  # +10% model bias

    # 1. SITE: modify first site (sin , cos waves)
    data[0, :, 0] += signal
    data[1, :, 0] += signal
    data[0, :36, 0] = np.nan  # invalidate first 1.5 days in obs

    meta = {
        "data_source": ["fakeobs", "fakemod"],
        "var_name": [var, var],
        "ts_type": "hourly",
        "filter_name": filter_name,
        "var_units": ["nmole mole-1", "nmole mole-1"],
        "min_num_obs": dict(hourly=dict(minutely=15), minutely=dict(secondly=15)),
    }

    meta.update(regfilter.to_dict())

    # create coordinates of DataArray
    coords = {
        "data_source": meta["data_source"],
        "time": dtime,
        "station_name": statnames,
        "latitude": ("station_name", lats),
        "longitude": ("station_name", lons),
        "altitude": ("station_name", alts),
    }

    dims = ["data_source", "time", "station_name"]
    cd = ColocatedData(data=data, coords=coords, dims=dims, name=var, attrs=meta)

    return cd


def _create_fake_partial_trends_coldata_3d(colocate_time: bool = False):
    var = "concpm10"
    filter_name = "ALL-wMOUNTAINS"
    regfilter = Filter(name=filter_name)

    dtime1 = pd.date_range("2000", "2001", freq="D")
    dtime2 = pd.date_range("2003", "2006", freq="D")
    dtime3 = pd.date_range("2006", "2009", freq="D")
    dtime4 = pd.date_range("2011", "2013", freq="D")
    dtime_partial = dtime1.union(dtime2).union(dtime3).union(dtime4)

    dtimes = [dtime1, dtime2, dtime_partial, dtime4, dtime_partial]

    lats = [-80, 0, 70, 0.1, 0]
    lons = [-150, 0, 100, 0.1, 3]
    alts = [0, 100, 2000, 10, 0]

    timenum = len(dtime_partial)
    statnum = len(lats)

    statnames = [f"FakeStation{c}" for c in range(statnum)]

    data = np.ones((2, timenum, statnum))

    for i in range(len(statnames)):
        data[1, :, i] = 1
        for j, d in enumerate(dtime_partial):
            if d in dtimes[i]:
                data[0, j, i] = 1
            else:
                data[0, j, i] = np.nan

    meta = {
        "data_source": ["fakeobs", "fakemod"],
        "var_name": [var, var],
        "ts_type": "monthly",
        "filter_name": filter_name,
        "ts_type_src": ["monthly", "daily"],
        "start_str": dtime_partial[0].strftime("%Y%m%d"),
        "stop_str": dtime_partial[-1].strftime("%Y%m%d"),
        "var_units": ["ug m-3", "ug m-3"],
        "vert_scheme": "surface",
        # "data_level": 3,
        # "revision_ref": "20210409",
        "from_files": [],
        "from_files_ref": None,
        "stations_ignored": None,
        "colocate_time": colocate_time,
        "obs_is_clim": False,
        "pyaerocom": "0.21.0.dev0",
        "min_num_obs": dict(monthly=dict(daily=15), daily=dict(hourly=12)),
        "resample_how": dict(monthly=dict(daily="sum"), daily=dict(hourly="max")),
    }

    meta.update(regfilter.to_dict())

    # create coordinates of DataArray
    coords = {
        "data_source": meta["data_source"],
        "time": dtime_partial,
        "station_name": statnames,
        "latitude": ("station_name", lats),
        "longitude": ("station_name", lons),
        "altitude": ("station_name", alts),
    }

    dims = ["data_source", "time", "station_name"]
    cd = ColocatedData(data=data, coords=coords, dims=dims, name=var, attrs=meta)

    return cd


def _create_fake_partial_trends_coldata_3d_coltime():
    return _create_fake_partial_trends_coldata_3d(True)


def _create_fake_partial_trends_coldata_3d_no_coltime():
    return _create_fake_partial_trends_coldata_3d()


def _create_fake_timeseries_hourly():
    cd = _create_fake_coldata_3d_hourly()
    obs = cd.data.data[0, :, 0]
    return pd.Series(obs, index=cd.data.time.data)


def _create_fake_coldata_4d():
    _lats_fake = np.arange(30, 60, 10)
    _lons_fake = np.arange(10, 30, 10)
    _time_fake = pd.date_range("2010-01", "2010-03", freq="MS")
    _data_fake = np.ones((2, len(_time_fake), len(_lats_fake), len(_lons_fake)))

    coords = {
        "data_source": ["obs", "mod"],
        "time": _time_fake,
        "latitude": _lats_fake,
        "longitude": _lons_fake,
    }

    dims = ["data_source", "time", "latitude", "longitude"]
    # set some obs vals NaN
    _data_fake[0, :, 1, 1] = np.nan
    _data_fake[0, 0, 0, 0] = np.nan
    meta = {"ts_type": "monthly"}
    return ColocatedData(data=_data_fake, coords=coords, dims=dims, attrs=meta)


COLDATA = dict(
    tm5_aeronet=lambda: ColocatedData(data=str(EXAMPLE_FILE)),
    fake_nodims=lambda: ColocatedData(data=np.ones((2, 1, 1))),
    fake_3d=_create_fake_coldata_3d,
    fake_4d=_create_fake_coldata_4d,
    fake_3d_hr=_create_fake_coldata_3d_hourly,
    fake_3d_trends=_create_fake_trends_coldata_3d,
    fake_3d_partial_trends_coltime=_create_fake_partial_trends_coldata_3d_coltime,
    fake_3d_partial_trends=_create_fake_partial_trends_coldata_3d_no_coltime,
)


@pytest.fixture
def coldata(coldataset: str) -> ColocatedData:
    """dispatch ColocatedData depending on coldataset"""
    return COLDATA[coldataset]()



from pydantic import BaseModel


class EvalRunOptions(BaseModel):
    clear_existing_json: bool = True
    only_json: bool = False
    only_colocation: bool = False
    #: If True, process only maps (skip obs evaluation)
    only_model_maps: bool = False
    obs_only: bool = False



from typing import Literal
from pydantic import (
    BaseModel,
    field_validator,
)
from pyaerocom.aeroval.exceptions import ConfigError
from pyaerocom.aeroval.helpers import (
    BoundingBox,
)

from pyaerocom.aeroval.modelmaps_helpers import CONTOUR, OVERLAY


PLOT_TYPE_OPTIONS = ({OVERLAY}, {CONTOUR}, {OVERLAY, CONTOUR})


class ModelMapsSetup(BaseModel):
    maps_freq: Literal["hourly", "daily", "monthly", "yearly", "coarsest"] = "coarsest"
    plot_types: dict[str, str | set[str]] | set[str] = {CONTOUR}
    boundaries: BoundingBox = BoundingBox(west=-180, east=180, north=90, south=-90)
    right_menu: tuple[str, ...] | None = None
    overlay_save_format: Literal["webp", "png"] = "webp"

    @field_validator("plot_types")
    def validate_plot_types(cls, v):
        if isinstance(v, dict):
            for m in v:
                if not isinstance(v[m], set):
                    v[m] = set([v[m]])  # v[m] must be a string
                if v[m] not in PLOT_TYPE_OPTIONS:
                    raise ConfigError("Model maps set up given a non-valid plot type.")
            return v
        if isinstance(v, str):
            v = set([v])
        if isinstance(v, list):  # can occur when reading a serialized config
            v = set(v)
        if v not in PLOT_TYPE_OPTIONS:
            raise ConfigError("Model maps set up given a non-valid plot type.")
        return v



import dataclasses

import numpy as np
import numpy.typing as npt

from pyaro.timeseries import (
    Reader,
    Data,
)

from pyaerocom.units.units_helpers import get_unit_conversion_fac
from pyaerocom.units.constants import M_N, M_S, M_O


@dataclasses.dataclass
class VariableScaling:
    REQ_VAR: str
    IN_UNIT: str
    OUT_UNIT: str
    SCALING_FACTOR: float
    OUT_VARNAME: str
    NOTE: str | None = None

    def required_input_variables(self) -> list[str]:
        return [self.REQ_VAR]

    def out_varname(self) -> str:
        return self.OUT_VARNAME


@dataclasses.dataclass
class VariableCombiner:
    REQ_VARS: tuple[str, str]
    IN_UNITS: tuple[str, str]
    OUT_UNIT: str
    OUT_VARNAME: str
    OP: str

    def required_input_variables(self) -> list[str]:
        return self.REQ_VARS

    def out_varname(self) -> str:
        return self.OUT_VARNAME


TRANSFORMATIONS = {
    "concNno_from_concno": VariableScaling(
        REQ_VAR="concno",
        IN_UNIT="ug m-3",
        OUT_UNIT="ug N m-3",
        SCALING_FACTOR=M_N / (M_N + M_O),
        OUT_VARNAME="concNno",
    ),
    "concNno2_from_concno2": VariableScaling(
        REQ_VAR="concno2",
        IN_UNIT="ug m-3",
        OUT_UNIT="ug N m-3",
        SCALING_FACTOR=M_N / (M_N + 2 * M_O),
        OUT_VARNAME="concNno2",
    ),
    "concSso2_from_concso2": VariableScaling(
        REQ_VAR="concso2",
        IN_UNIT="ug m-3",
        OUT_UNIT="ug S m-3",
        SCALING_FACTOR=M_S / (M_S + 2 * M_O),
        OUT_VARNAME="concSso2",
    ),
    "vmro3_from_conco3": VariableScaling(
        REQ_VAR="conco3",
        IN_UNIT="µg m-3",
        OUT_UNIT="ppb",
        SCALING_FACTOR=0.5011,  # 20C and 1013 hPa
        OUT_VARNAME="vmro3",
        NOTE="The vmro3_from_conco3 transform is only valid at T=20C, p=1013hPa",
    ),
    "vmro3max_from_conco3": VariableScaling(  # Requires `resample_how`
        REQ_VAR="conco3",
        IN_UNIT="µg m-3",
        OUT_UNIT="ppb",
        SCALING_FACTOR=0.5011,  # 20C and 1013 hPa
        OUT_VARNAME="vmro3max",
        NOTE="The vmro3max_from_conco3 transform is only valid at T=20C, p=1013hPa, and the transform requires the use of resample_how to obtain the daily maximum",
    ),
    "vmrno2_from_concno2": VariableScaling(
        REQ_VAR="concno2",
        IN_UNIT="ug m-3",
        OUT_UNIT="ppb",
        SCALING_FACTOR=0.5229,  # 20C and 1013 hPa
        OUT_VARNAME="vmrno2",
        NOTE="The vmrno2_from_conco3 transform is only valid at T=20C, p=1013hPa",
    ),
    "vmrox_from_vmrno2_vmro3": VariableCombiner(
        REQ_VARS=("vmrno2", "vmro3"),
        IN_UNITS=("nmol mol-1", "nmol mol-1"),
        OUT_UNIT="nmol mol-1",
        OUT_VARNAME="vmrox",
        OP="ADD",
    ),
}


class PostProcessingReaderData(Data):
    def __init__(self, data: Data, variable: str, units: str, scaling: float | None):
        self._variable = variable
        self._units = units
        self.data = data
        self.scaling = scaling

    def keys(self):
        return self.data.keys()

    def slice(self, index):
        return PostProcessingReaderData(
            self.data.slice(index),
            variable=self._variable,
            units=self._units,
            scaling=self.scaling,
        )

    def __len__(self):
        return self.data.__len__()

    @property
    def variable(self) -> str:
        return self._variable

    @property
    def units(self) -> str:
        return self._units

    @property
    def values(self):
        if self.scaling is None:
            return self.data.values
        else:
            return self.data.values * self.scaling

    @property
    def stations(self):
        return self.data.stations

    @property
    def latitudes(self):
        return self.data.latitudes

    @property
    def longitudes(self):
        return self.data.longitudes

    @property
    def altitudes(self):
        return self.data.altitudes

    @property
    def start_times(self):
        return self.data.start_times

    @property
    def end_times(self):
        return self.data.end_times

    @property
    def flags(self):
        return self.data.flags

    @property
    def standard_deviations(self):
        if self.scaling is None:
            return self.data.standard_deviations
        else:
            return self.data.standard_deviations * self.scaling


class PostProcessingReaderException(Exception):
    pass


class PostProcessingReader(Reader):
    def __init__(
        self,
        reader: Reader,
        compute_vars: list[str] | None = None,
    ):
        self.reader = reader

        self.compute_vars = dict()
        if compute_vars is not None:
            known_variables = reader.variables()
            for compute_var in compute_vars:
                transform = TRANSFORMATIONS.get(compute_var)
                if transform is None:
                    raise PostProcessingReaderException(
                        f"Unknown transformation ({compute_var}) encountered"
                    )
                required_input = transform.required_input_variables()
                missing = set(required_input) - set(known_variables)
                if len(missing) > 0:
                    raise PostProcessingReaderException(
                        f"The transformation {compute_var} requires variables which are not present, missing {missing}"
                    )
                known_variables.append(transform.out_varname())
                self.compute_vars[transform.out_varname()] = transform

    def metadata(self) -> dict[str, str]:
        return self.reader.metadata()

    def data(self, varname: str) -> Data:
        if varname not in self.compute_vars:
            data = self.reader.data(varname)
            return data
        transform = self.compute_vars[varname]
        if isinstance(transform, VariableScaling):
            data = self.reader.data(transform.REQ_VAR)
            scaling = transform.SCALING_FACTOR * get_unit_conversion_fac(
                from_unit=data.units,
                to_unit=transform.IN_UNIT,
                var_name=transform.REQ_VAR,
            )
            return PostProcessingReaderData(
                data, variable=varname, units=transform.OUT_UNIT, scaling=scaling
            )
        if isinstance(transform, VariableCombiner):
            data = [self.data(var) for var in transform.REQ_VARS]
            scalings = [
                get_unit_conversion_fac(from_unit=d.units, to_unit=out_unit)
                for d, out_unit in zip(data, transform.IN_UNITS)
            ]

            # Find unique shared data based on lat/lon and times
            groupbys = [np.array((d.latitudes, d.longitudes)) for d in data]
            uniqs = [
                set(map(tuple, np.unique(group, axis=1).transpose().tolist()))
                for group in groupbys
            ]
            shared = np.array(list(set.intersection(*uniqs)))

            new_latitudes = []
            new_longitudes = []
            new_starttimes = []
            new_endtimes = []
            new_stations = []
            new_altitudes = []
            new_values = []

            for lat, lon in shared:  # Per station
                masks = [(d.latitudes == lat) & (d.longitudes == lon) for d in data]
                data_subset = [d[mask] for d, mask in zip(data, masks)]

                start_times = [d.start_times for d in data_subset]
                indexings = [np.argsort(s) for s in start_times]

                start_times = [s[i] for s, i in zip(start_times, indexings)]
                end_times = [d.end_times[i] for d, i in zip(data_subset, indexings)]
                stations = data_subset[0].stations[indexings[0]]
                altitudes = data_subset[0].altitudes[indexings[0]]

                try:
                    lindex, rindex = matching_indices(start_times[0], start_times[1])
                except ValueError:
                    continue

                if not np.all(end_times[0][lindex] == end_times[1][rindex]):
                    continue  # Different durations encountered, skip this station

                new_latitudes.append(np.full(len(lindex), fill_value=lat))
                new_longitudes.append(np.full(len(lindex), fill_value=lon))
                new_starttimes.append(start_times[0][lindex])
                new_endtimes.append(start_times[0][lindex])
                new_stations.append(stations[lindex])
                new_altitudes.append(altitudes[lindex])

                if transform.OP == "ADD":
                    values = (
                        data_subset[0].values[lindex] * scalings[0]
                        + data_subset[1].values[rindex] * scalings[1]
                    )
                else:
                    raise PostProcessingReaderException(
                        f"Transform mode {transform.OP} is not supported"
                    )
                new_values.append(values)

            newdata = {
                "latitudes": np.concatenate(new_latitudes),
                "longitudes": np.concatenate(new_longitudes),
                "start_times": np.concatenate(new_starttimes),
                "end_times": np.concatenate(new_endtimes),
                "stations": np.concatenate(new_stations),
                "altitudes": np.concatenate(new_altitudes),
                "values": np.concatenate(new_values),
            }

            n = len(newdata["latitudes"])
            newdata.update(
                {
                    "standard_deviations": np.full(n, fill_value=np.nan),
                    "flags": np.ones(n),
                }
            )

            return DictBackedData(
                newdata, variable=transform.out_varname(), units=transform.OUT_UNIT
            )
        else:
            raise PostProcessingReaderException(
                f"Unknown transform {transform} encountered for variable {varname}"
            )

    def variables(self) -> list[str]:
        variables = list()
        variables.extend(self.reader.variables())
        variables.extend(self.compute_vars.keys())
        return variables

    def stations(self):
        return self.reader.stations()

    def close(self) -> None:
        self.reader.close()


def matching_indices(x, y) -> tuple[npt.NDArray[int], npt.NDArray[int]]:
    """Returns indices of x and y such that x[xind] == y[yind]
    x and y must be monotonically increasing
    """
    dx = np.diff(x)
    # Use numpy.zeros to get a zero of the same dtype
    if not np.all(dx > np.zeros(1, dtype=dx.dtype)):
        raise ValueError("x is not monotonically increasing")
    dy = np.diff(y)
    if not np.all(dy > np.zeros(1, dtype=dy.dtype)):
        raise ValueError("y is not monotonically increasing")
    x_indices, y_indices = list(), list()
    ix, iy = 0, 0
    while (ix < len(x)) and (iy < len(y)):
        le = x[ix]
        ri = y[iy]
        if le == ri:
            x_indices.append(ix)
            y_indices.append(iy)
            ix += 1
            iy += 1
        elif le < ri:
            ix += 1
        else:
            iy += 1

    return x_indices, y_indices


class DictBackedData(Data):
    def __init__(self, data, variable: str, units: str):
        self._variable = variable
        self._units = units
        self._data = data

    def keys(self):
        return {}.keys()

    def slice(self, index):
        return DictBackedData(
            data=self._data()[index],
            variable=self.variable,
            units=self._units,
        )

    def __len__(self):
        return len(self.values)

    @property
    def variable(self) -> str:
        return self._variable

    @property
    def units(self) -> str:
        return self._units

    @property
    def values(self):
        return self._data["values"]

    @property
    def stations(self):
        return self._data["stations"]

    @property
    def latitudes(self):
        return self._data["latitudes"]

    @property
    def longitudes(self):
        return self._data["longitudes"]

    @property
    def altitudes(self):
        return self._data["altitudes"]

    @property
    def start_times(self):
        return self._data["start_times"]

    @property
    def end_times(self):
        return self._data["end_times"]

    @property
    def flags(self):
        return self._data["flags"]

    @property
    def standard_deviations(self):
        return self._data["standard_deviations"]
